# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from fractions import Fraction

import numpy as np
import sympy


def Q_to_sympy(x: float) -> sympy.Expr:
    x = Fraction(x).limit_denominator()
    return sympy.sympify(x)


def sqrtQ_to_sympy(x: float) -> sympy.Expr:
    sign = 1 if x >= 0 else -1
    return sign * sympy.sqrt(Q_to_sympy(x**2))


def sqrtQarray_to_sympy(x: np.ndarray) -> sympy.Array:
    if x.ndim == 0:
        return sqrtQ_to_sympy(x)
    else:
        return sympy.Array([sqrtQarray_to_sympy(row) for row in x])



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .rep import Rep
from .irrep import (
    Irrep,
    clebsch_gordan,
    selection_rule_product,
    selection_rule_power,
)
from .irrep_su2 import SU2
from .irrep_so3 import SO3
from .irrep_o3 import O3


__all__ = [
    "Rep",
    "Irrep",
    "clebsch_gordan",
    "selection_rule_product",
    "selection_rule_power",
    "SU2",
    "SO3",
    "O3",
]



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses
from typing import Any, Callable

import numpy as np

import cuequivariance as cue
from cuequivariance.segmented_polynomials.segmented_tensor_product import (
    _canonicalize_index,
)


@dataclasses.dataclass(init=False, frozen=True)
class EquivariantPolynomial:
    """A polynomial representation with equivariance constraints.

    This class extends :class:`cue.SegmentedPolynomial <cuequivariance.SegmentedPolynomial>` by incorporating information about the group representations
    of each input and output tensor. It ensures that operations performed by the polynomial respect
    the equivariance constraints defined by these representations, making it suitable for building
    equivariant neural networks.

    Args:
        inputs (tuple of :class:`cue.Rep <cuequivariance.Rep>`): Group representations for input tensors.
        outputs (tuple of :class:`cue.Rep <cuequivariance.Rep>`): Group representations for output tensors.
        polynomial (:class:`cue.SegmentedPolynomial <cuequivariance.SegmentedPolynomial>`): The underlying polynomial transformation.
    """

    inputs: tuple[cue.Rep, ...]
    outputs: tuple[cue.Rep, ...]
    polynomial: cue.SegmentedPolynomial

    # ------------------------------------------------------------------------
    # Core Structure and Initialization
    # ------------------------------------------------------------------------

    def __init__(
        self,
        inputs: list[cue.Rep],
        outputs: list[cue.Rep],
        polynomial: cue.SegmentedPolynomial,
    ):
        assert isinstance(polynomial, cue.SegmentedPolynomial)
        object.__setattr__(self, "inputs", tuple(inputs))
        object.__setattr__(self, "outputs", tuple(outputs))
        object.__setattr__(self, "polynomial", polynomial)
        if len(self.inputs) != self.polynomial.num_inputs:
            raise ValueError(
                f"Number of inputs {len(self.inputs)} must equal the number of inputs"
                f" in the polynomial {self.polynomial.num_inputs}"
            )
        if len(self.outputs) != self.polynomial.num_outputs:
            raise ValueError(
                f"Number of outputs {len(self.outputs)} must equal the number of outputs"
                f" in the polynomial {self.polynomial.num_outputs}"
            )
        for rep, ope in zip(self.operands, self.polynomial.operands):
            assert ope.size == rep.dim, (
                f"{ope} incompatible with {rep}. {ope.size=} != {rep.dim=}"
            )

    @property
    def operands(self) -> tuple[cue.Rep, ...]:
        """Get all operands (inputs and outputs) of the polynomial.

        Returns:
            tuple of :class:`cue.Rep <cuequivariance.Rep>`: Tuple of all operands.
        """
        return self.inputs + self.outputs

    @property
    def num_inputs(self) -> int:
        """Get the number of input operands.

        Returns:
            int: Number of input operands.
        """
        return len(self.inputs)

    @property
    def num_outputs(self) -> int:
        """Get the number of output operands.

        Returns:
            int: Number of output operands.
        """
        return len(self.outputs)

    @property
    def num_operands(self) -> int:
        """Get the total number of operands (inputs + outputs).

        Returns:
            int: Total number of operands.
        """
        return self.num_inputs + self.num_outputs

    # ------------------------------------------------------------------------
    # Class Construction Methods
    # ------------------------------------------------------------------------

    # Method eval_last_operand is present in SegmentedPolynomial but not in EquivariantPolynomial

    @classmethod
    def stack(
        cls, polys: list[EquivariantPolynomial], stacked: list[bool]
    ) -> EquivariantPolynomial:
        """Stack equivariant polynomials together.

        This method combines multiple polynomials by stacking their operands where indicated
        by the stacked parameter. Non-stacked operands must be identical across all polynomials.

        Args:
            polys (list of :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`): List of polynomials to stack.
            stacked (list of bool): List indicating which operands should be stacked.

        Returns:
            :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`: The stacked polynomial.

        Example:
            >>> p1 = cue.descriptors.spherical_harmonics(cue.SO3(1), [1, 2])
            >>> p2 = cue.descriptors.spherical_harmonics(cue.SO3(1), [2, 3])
            >>> cue.EquivariantPolynomial.stack([p1, p2], [False, True])
            ╭ a=1 -> B=1+2+2+3
            │  []·a[]➜B[] ───────── num_paths=3
            │  []·a[]·a[]➜B[] ───── num_paths=22
            ╰─ []·a[]·a[]·a[]➜B[] ─ num_paths=41

            Note how the STPs of degree 2 has been automatically fused into a single STP.
        """
        assert len(polys) > 0
        num_operands = polys[0].num_operands

        assert all(pol.num_operands == num_operands for pol in polys)
        assert len(stacked) == num_operands

        operands = []
        for oid in range(num_operands):
            if stacked[oid]:
                for pol in polys:
                    if not isinstance(pol.operands[oid], cue.IrrepsAndLayout):
                        raise ValueError(
                            f"Cannot stack operand {oid} of type {type(pol.operands[oid])}"
                        )
                operands.append(cue.concatenate([pol.operands[oid] for pol in polys]))
            else:
                ope = polys[0].operands[oid]
                for pol in polys:
                    if pol.operands[oid] != ope:
                        raise ValueError(
                            f"Operand {oid} must be the same for all polynomials."
                            f" Found {ope} and {pol.operands[oid]}"
                        )
                operands.append(ope)

        p = cue.SegmentedPolynomial.stack([pol.polynomial for pol in polys], stacked)
        return cls(operands[: p.num_inputs], operands[p.num_inputs :], p)

    # Method stack_tensor_products is present in SegmentedPolynomial but not in EquivariantPolynomial

    # Method concatenate is present in SegmentedPolynomial but not in EquivariantPolynomial

    # ------------------------------------------------------------------------
    # Standard Python Methods
    # ------------------------------------------------------------------------

    def __repr__(self):
        return self.polynomial.to_string([f"{rep}" for rep in self.operands])

    def __call__(self, *inputs: np.ndarray) -> list[np.ndarray]:
        """Evaluate the polynomial on the given inputs.

        Args:
            *inputs (np.ndarray): Input arrays to evaluate the polynomial on.

        Returns:
            list of np.ndarray: List of output arrays.

        Note:
            This is a reference implementation using numpy and may not be optimized for performance.
        """
        return self.polynomial(*inputs)

    def __hash__(self) -> int:
        return hash((self.inputs, self.outputs, self.polynomial))

    def __eq__(self, value) -> bool:
        assert isinstance(value, EquivariantPolynomial)
        return (
            self.inputs == value.inputs
            and self.outputs == value.outputs
            and self.polynomial == value.polynomial
        )

    def __lt__(self, value) -> bool:
        assert isinstance(value, EquivariantPolynomial)
        return (
            self.inputs,
            self.outputs,
            self.polynomial,
        ) < (
            value.inputs,
            value.outputs,
            value.polynomial,
        )

    def __mul__(self, factor: float) -> EquivariantPolynomial:
        return EquivariantPolynomial(
            self.inputs,
            self.outputs,
            self.polynomial * factor,
        )

    def __rmul__(self, factor: float) -> EquivariantPolynomial:
        return self.__mul__(factor)

    # ------------------------------------------------------------------------
    # Analysis Methods
    # ------------------------------------------------------------------------

    def all_same_segment_shape(self) -> bool:
        """Check if all operands have the same segment shape.

        Returns:
            bool: True if all operands have the same segment shape.
        """
        return self.polynomial.all_same_segment_shape()

    def used_inputs(self) -> list[bool]:
        """Get list of boolean values indicating which inputs are used in the polynomial.

        Returns:
            list of bool: List where True indicates the input is used.
        """
        return self.polynomial.used_inputs()

    def used_outputs(self) -> list[bool]:
        """Get list of boolean values indicating which outputs are used in the polynomial.

        Returns:
            list of bool: List where True indicates the output is used.
        """
        return self.polynomial.used_outputs()

    def used_operands(self) -> list[bool]:
        """Get list of boolean values indicating which operands are used in the polynomial.

        Returns:
            list of bool: List where True indicates the operand is used.
        """
        return self.polynomial.used_operands()

    def flop(self, batch_size: int = 1) -> int:
        """Compute the number of floating point operations in the polynomial.

        Args:
            batch_size (int, optional): Batch size for computation. Defaults to 1.

        Returns:
            int: Number of floating point operations.
        """
        return self.polynomial.flop(batch_size)

    def memory(self, batch_sizes: list[int]) -> int:
        """Compute the memory usage of the polynomial.

        Args:
            batch_sizes (list of int): List of batch sizes for each operand. Each operand
                can have its own batch size, allowing for different batch dimensions
                per tensor.

        Returns:
            int: Memory usage in number of elements.
        """
        assert len(batch_sizes) == len(self.operands)
        return sum(Z * rep.dim for Z, rep in zip(batch_sizes, self.operands))

    # ------------------------------------------------------------------------
    # Transformation Methods
    # ------------------------------------------------------------------------

    def apply_fn(
        self,
        f: Callable[
            [cue.Operation, cue.SegmentedTensorProduct],
            tuple[cue.Operation, cue.SegmentedTensorProduct] | None,
        ],
    ) -> EquivariantPolynomial:
        """Apply a function to each tensor product in the polynomial.

        Args:
            f (Callable): Function to apply to each operation and tensor product pair.

        Returns:
            :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`: New polynomial with transformed tensor products.
        """
        new_polynomial = self.polynomial.apply_fn(f)
        return EquivariantPolynomial(self.inputs, self.outputs, new_polynomial)

    def fuse_stps(self) -> EquivariantPolynomial:
        """Fuse segmented tensor products with identical operations and operands.

        Returns:
            :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`: Polynomial with fused tensor products.
        """
        return EquivariantPolynomial(
            self.inputs, self.outputs, self.polynomial.fuse_stps()
        )

    def consolidate(self) -> EquivariantPolynomial:
        """Consolidate the segmented tensor products by removing empty segments and squeezing modes.

        Returns:
            :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`: Consolidated polynomial.
        """
        return EquivariantPolynomial(
            self.inputs, self.outputs, self.polynomial.consolidate()
        )

    def flatten_modes(self, modes: list[str]) -> EquivariantPolynomial:
        """Flatten specified modes in the polynomial.

        Args:
            modes (list of str): List of mode names to flatten.

        Returns:
            :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`: Polynomial with flattened modes.
        """
        return EquivariantPolynomial(
            self.inputs, self.outputs, self.polynomial.flatten_modes(modes)
        )

    def canonicalize_subscripts(self) -> EquivariantPolynomial:
        """Canonicalize the subscripts of the segmented tensor products.

        Returns:
            :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`: Polynomial with canonicalized subscripts.
        """
        return EquivariantPolynomial(
            self.inputs, self.outputs, self.polynomial.canonicalize_subscripts()
        )

    def squeeze_modes(self, modes: str | None = None) -> EquivariantPolynomial:
        """Squeeze specified modes in the polynomial.

        Args:
            modes (str | None, optional): Modes to squeeze. If None, squeezes all modes.

        Returns:
            :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`: Polynomial with squeezed modes.
        """
        return EquivariantPolynomial(
            self.inputs, self.outputs, self.polynomial.squeeze_modes(modes)
        )

    def split_mode(self, mode: str, size: int) -> EquivariantPolynomial:
        """Split specified mode in the polynomial.

        Args:
            mode (str): Mode to split.
            size (int): Size to split the mode into.

        Returns:
            :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`: Polynomial with split mode.
        """
        return EquivariantPolynomial(
            self.inputs, self.outputs, self.polynomial.split_mode(mode, size)
        )

    def flatten_coefficient_modes(self) -> EquivariantPolynomial:
        """Flatten the coefficient modes of the segmented tensor products.

        Returns:
            :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`: Polynomial with flattened coefficient modes.
        """
        return EquivariantPolynomial(
            self.inputs, self.outputs, self.polynomial.flatten_coefficient_modes()
        )

    def symmetrize_for_identical_operands(self) -> EquivariantPolynomial:
        """Symmetrize the paths of the segmented tensor products for identical operands.

        This operation increases the number of paths in the segmented tensor products.

        Returns:
            :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`: Polynomial with symmetrized paths.
        """
        return EquivariantPolynomial(
            self.inputs,
            self.outputs,
            self.polynomial.symmetrize_for_identical_operands(),
        )

    def unsymmetrize_for_identical_operands(self) -> EquivariantPolynomial:
        """Unsymmetrize the paths of the segmented tensor products for identical operands.

        This operation decreases the number of paths in the segmented tensor products.

        Returns:
            :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`: Polynomial with unsymmetrized paths.
        """
        return EquivariantPolynomial(
            self.inputs,
            self.outputs,
            self.polynomial.unsymmetrize_for_identical_operands(),
        )

    def split_operand_by_irrep(self, operand_id: int) -> EquivariantPolynomial:
        """Split an irreps operand into separate operands for each (mul, ir) pair.

        Args:
            operand_id (int): Index of the operand to split.

        Returns:
            EquivariantPolynomial: New polynomial with the specified operand split by irreps.

        Raises:
            AssertionError: If the operand is not an IrrepsAndLayout instance.

        Example:
            >>> e = cue.descriptors.channelwise_tensor_product(
            ...     cue.Irreps(cue.SO3, "64x0 + 32x1"), cue.Irreps(cue.SO3, "0 + 1"), simplify_irreps3=True
            ... )
            >>> e.split_operand_by_irrep(-1)
            ╭ a=256x0 b=64x0+32x1 c=0+1 -> D=96x0 E=128x1 F=32x2
            │  []·a[u]·b[u]·c[]➜D[u] ─ num_paths=4 u={32, 64}
            │  []·a[u]·b[u]·c[]➜E[u] ─ num_paths=12 u={32, 64}
            ╰─ []·a[u]·b[u]·c[]➜F[u] ─ num_paths=11 u={32, 64}
        """
        operand_id = _canonicalize_index("operand_id", operand_id, self.num_operands)
        operand = self.operands[operand_id]
        assert isinstance(operand, cue.IrrepsAndLayout)

        # Create polynomial slices for each irrep
        offsets = [0]
        i = 0
        for mul, ir in operand.irreps:
            i += mul * ir.dim
            offsets.append(i)

        poly = self.polynomial.split_operand_by_size(operand_id, offsets)

        # # Calculate new operand counts and create mapping function
        num_new = len(offsets) - 1
        new_num_inputs = self.num_inputs + (
            num_new - 1 if operand_id < self.num_inputs else 0
        )

        # Create final operands structure
        splits = tuple(
            cue.IrrepsAndLayout(operand.irreps[i : i + 1], operand.layout)
            for i in range(len(operand.irreps))
        )
        new_operands = (
            self.operands[:operand_id] + splits + self.operands[operand_id + 1 :]
        )

        return EquivariantPolynomial(
            new_operands[:new_num_inputs], new_operands[new_num_inputs:], poly
        )

    # ------------------------------------------------------------------------
    # Filtering Methods
    # ------------------------------------------------------------------------

    def filter_keep_operands(self, keep: list[bool]) -> EquivariantPolynomial:
        """Select which operands to keep in the polynomial.

        Use this method when you want to compute only a subset of the polynomial outputs
        and have control over which inputs to keep. For keeping all inputs (even if
        not used), use filter_keep_outputs. For automatically removing unused operands,
        use filter_drop_unsued_operands.

        Args:
            keep (list of bool): List indicating which operands to keep.

        Returns:
            :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`: Polynomial with selected operands.
        """
        assert len(keep) == self.num_operands

        filtered_polynomial = self.polynomial.filter_keep_operands(keep)

        filtered_inputs = [
            rep for rep, k in zip(self.inputs, keep[: self.num_inputs]) if k
        ]
        filtered_outputs = [
            rep for rep, k in zip(self.outputs, keep[self.num_inputs :]) if k
        ]

        return EquivariantPolynomial(
            filtered_inputs, filtered_outputs, filtered_polynomial
        )

    def filter_keep_outputs(self, keep: list[bool]) -> EquivariantPolynomial:
        """Select which outputs to keep in the polynomial.

        Args:
            keep (list of bool): List indicating which outputs to keep.

        Returns:
            :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`: Polynomial with selected outputs.
        """
        assert len(keep) == self.num_outputs
        return self.filter_keep_operands([True] * self.num_inputs + keep)

    def filter_drop_unsued_operands(self) -> EquivariantPolynomial:
        """Remove all unused operands from the polynomial.

        Returns:
            :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`: Polynomial with unused operands removed.
        """
        used = self.used_operands()

        filtered_inputs = [
            rep
            for rep, used_flag in zip(self.inputs, used[: self.num_inputs])
            if used_flag
        ]
        filtered_outputs = [
            rep
            for rep, used_flag in zip(self.outputs, used[self.num_inputs :])
            if used_flag
        ]

        filtered_polynomial = self.polynomial.filter_drop_unsued_operands()

        return EquivariantPolynomial(
            filtered_inputs, filtered_outputs, filtered_polynomial
        )

    def compute_only(self, keep: list[bool]) -> EquivariantPolynomial:
        """Create a polynomial that only computes selected outputs.

        The new polynomial will keep the same operands as the original one,
        but will only compute the selected outputs.

        Args:
            keep (list of bool): List indicating which outputs to compute.

        Returns:
            :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`: Polynomial computing only selected outputs.
        """
        assert len(keep) == self.num_outputs
        filtered_polynomial = self.polynomial.compute_only(keep)
        return EquivariantPolynomial(self.inputs, self.outputs, filtered_polynomial)

    # ------------------------------------------------------------------------
    # Automatic Differentiation Methods
    # ------------------------------------------------------------------------

    def jvp(
        self, has_tangent: list[bool]
    ) -> tuple[
        EquivariantPolynomial,
        Callable[[tuple[list[Any], list[Any]]], tuple[list[Any], list[Any]]],
    ]:
        """Compute the Jacobian-vector product of the polynomial.

        Args:
            has_tangent (list of bool): List indicating which inputs have tangents.

        Returns:
            tuple of :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>` and Callable:
                The JVP polynomial and a mapping function for inputs/outputs.
        """
        p, m = self.polynomial.jvp(has_tangent)
        return EquivariantPolynomial(*m((self.inputs, self.outputs)), p), m

    def transpose(
        self,
        is_undefined_primal: list[bool],
        has_cotangent: list[bool],
    ) -> tuple[
        EquivariantPolynomial,
        Callable[[tuple[list[Any], list[Any]]], tuple[list[Any], list[Any]]],
    ]:
        """Transpose the polynomial for reverse-mode automatic differentiation.

        Args:
            is_undefined_primal (list of bool): List indicating which inputs have undefined primals.
            has_cotangent (list of bool): List indicating which outputs have cotangents.

        Returns:
            tuple of :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>` and Callable:
                The transposed polynomial and a mapping function for inputs/outputs.
        """
        p, m = self.polynomial.transpose(is_undefined_primal, has_cotangent)
        return EquivariantPolynomial(*m((self.inputs, self.outputs)), p), m

    def backward(
        self, requires_gradient: list[bool], has_cotangent: list[bool]
    ) -> tuple[
        EquivariantPolynomial,
        Callable[[tuple[list[Any], list[Any]]], tuple[list[Any], list[Any]]],
    ]:
        """Compute the backward pass of the polynomial for gradient computation.

        Args:
            requires_gradient (list of bool): List indicating which inputs require gradients.
            has_cotangent (list of bool): List indicating which outputs have cotangents.

        Returns:
            tuple of :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>` and Callable:
                The backward polynomial and a mapping function for inputs/outputs.
        """
        p, m = self.polynomial.backward(requires_gradient, has_cotangent)
        return EquivariantPolynomial(*m((self.inputs, self.outputs)), p), m



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def format_set(s: set[int]) -> str:
    if len(s) == 0:
        return ""
    if len(s) == 1:
        return str(next(iter(s)))
    return "{" + ", ".join(str(i) for i in sorted(s)) + "}"


def format_dimensions_dict(dims: dict[str, set[int]]) -> str:
    return " ".join(f"{m}={format_set(dd)}" for m, dd in sorted(dims.items()))



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import cuequivariance.segmented_polynomials as sp
import pytest


def test_subscripts():
    with pytest.raises(ValueError):
        sp.Subscripts("#$%@")

    with pytest.raises(ValueError):
        sp.Subscripts("Zu")  # uppercase not supported anymore

    with pytest.raises(ValueError):
        sp.Subscripts("uZ")  # uppercase after lowercase

    with pytest.raises(ValueError):
        sp.Subscripts("uZ+ij+kl")  # multiple + signs

    subscripts = sp.Subscripts("ui,vj,uvk+ijk")
    assert subscripts.canonicalize() == "ui,vj,uvk+ijk"

    assert subscripts.coefficients == "ijk"

    assert subscripts.num_operands == 3
    assert subscripts.operands[0] == "ui"
    assert subscripts.operands[1] == "vj"
    assert subscripts.operands[2] == "uvk"

    assert subscripts.is_subset_of("uwi,vwj,uvk+ijk")  # using w=1
    assert subscripts.is_equivalent("ui_vj_uvk+ijk")


def test_canonicalize():
    assert sp.Subscripts("ui").canonicalize() == "uv"
    assert sp.Subscripts("ab,ad+bd").canonicalize() == "ui,uj+ij"
    assert sp.Subscripts("i,j+ji").canonicalize() == "i,j+ji"
    assert sp.Subscripts("j,i+ij").canonicalize() == "i,j+ji"



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import itertools
import re
from dataclasses import dataclass
from typing import Iterator

import numpy as np

from cuequivariance.etc.linalg import round_to_sqrt_rational
from cuequivariance.group_theory.representations import SU2, Irrep


# This function is copied from https://github.com/lie-nn/lie-nn/blob/70adebce44e3197ee17f780585c6570d836fc2fe/lie_nn/_src/irreps/so3_real.py
def change_basis_real_to_complex(ell: int) -> np.ndarray:
    # https://en.wikipedia.org/wiki/Spherical_harmonics#Real_form
    q = np.zeros((2 * ell + 1, 2 * ell + 1), dtype=np.complex128)
    for m in range(-ell, 0):
        q[ell + m, ell + abs(m)] = 1 / np.sqrt(2)
        q[ell + m, ell - abs(m)] = -1j / np.sqrt(2)
    q[ell, ell] = 1
    for m in range(1, ell + 1):
        q[ell + m, ell + abs(m)] = (-1) ** m / np.sqrt(2)
        q[ell + m, ell - abs(m)] = 1j * (-1) ** m / np.sqrt(2)

    # Added factor of 1j**l to make the Clebsch-Gordan coefficients real
    return q * (-1j) ** ell


# This class is adapted from https://github.com/lie-nn/lie-nn/blob/70adebce44e3197ee17f780585c6570d836fc2fe/lie_nn/_src/irreps/so3_real.py
@dataclass(frozen=True)
class SO3(Irrep):
    r"""Subclass of :class:`Irrep`, real irreducible representations of the 3D rotation group :math:`SO(3)`.

    Each representation is labeled by a non-negative integer :math:`l`.

    Examples:
        >>> SO3(0)
        0
        >>> SO3(1)
        1
        >>> SO3(1).dim
        3
        >>> SO3.from_string("2")
        2
    """

    l: int  # noqa: E741

    @classmethod
    def regexp_pattern(cls) -> re.Pattern:
        return re.compile(r"(\d+)")

    @classmethod
    def from_string(cls, s: str) -> SO3:
        return cls(l=int(s))

    def __repr__(rep: SO3) -> str:
        return f"{rep.l}"

    def __mul__(rep1: SO3, rep2: SO3) -> Iterator[SO3]:
        rep2 = rep1._from(rep2)
        return [
            SO3(l=ell) for ell in range(abs(rep1.l - rep2.l), rep1.l + rep2.l + 1, 1)
        ]

    @classmethod
    def clebsch_gordan(cls, rep1: SO3, rep2: SO3, rep3: SO3) -> np.ndarray:
        rep1, rep2, rep3 = cls._from(rep1), cls._from(rep2), cls._from(rep3)

        # return an array of shape ``(number_of_paths, rep1.dim, rep2.dim, rep3.dim)``
        C = SU2.clebsch_gordan(SU2(j=rep1.l), SU2(j=rep2.l), SU2(j=rep3.l))
        Q1 = change_basis_real_to_complex(rep1.l)
        Q2 = change_basis_real_to_complex(rep2.l)
        Q3 = change_basis_real_to_complex(rep3.l)
        C = np.einsum("ij,kl,mn,zikn->zjlm", Q1, Q2, np.conj(Q3.T), C)

        # ensure it's real
        assert np.all(np.abs(np.imag(C)) < 1e-5)
        C = np.real(C)

        return C

    @property
    def dim(rep: SO3) -> int:
        return 2 * rep.l + 1

    def __lt__(rep1: SO3, rep2: SO3) -> bool:
        rep2 = rep1._from(rep2)
        return rep1.l < rep2.l

    @classmethod
    def iterator(cls) -> Iterator[SO3]:
        for ell in itertools.count(0):
            yield cls(l=ell)

    def continuous_generators(rep: SO3) -> np.ndarray:
        X = SU2(j=rep.l).X
        Q = change_basis_real_to_complex(rep.l)
        X = np.conj(Q.T) @ X @ Q
        assert np.all(np.abs(np.imag(X)) < 1e-5)
        return np.real(X)

    def discrete_generators(rep: SO3) -> np.ndarray:
        return np.zeros((0, rep.dim, rep.dim))

    def algebra(rep=None) -> np.ndarray:
        # [X_i, X_j] = A_ijk X_k
        return SU2.algebra(rep)

    def rotation(self, axis: np.ndarray, angle: float) -> np.ndarray:
        """Rotation matrix for the representation."""
        ell = self.l
        m = np.arange(-ell, ell + 1)

        axis = axis / np.linalg.norm(axis)
        iX = 1j * np.sum(self.X * axis[:, None, None], axis=0)
        _val, V = np.linalg.eigh(iX)

        # np.testing.assert_allclose(_val, m, atol=1e-10)
        # np.testing.assert_allclose(V @ np.diag(m) @ V.T.conj(), iX, atol=1e-10)

        phase = np.exp(-1j * angle * m)

        R = V @ np.diag(phase) @ V.T.conj()
        # np.testing.assert_allclose(R.imag, 0, atol=1e-10)
        R = R.real

        if abs(angle) in [0.0, np.pi / 2, np.pi, 3 * np.pi / 2]:
            R = round_to_sqrt_rational(R, 2**40 + 1)

        return R

    # def exp_map(
    #     self, continuous_params: np.ndarray, discrete_params: np.ndarray
    # ) -> np.ndarray:
    #     axis = continuous_params
    #     angle = np.linalg.norm(axis)
    #     if angle == 0:
    #         return np.eye(self.dim)

    #     return self.rotation(axis, angle)



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

import cuequivariance as cue


def test_su2():
    irreps = cue.Irreps("SU2", "0 + 2x1/2 + 1")
    assert len(irreps) == 3

    assert irreps[0].mul == 1
    assert irreps[0].ir.j == 0

    assert irreps[1].mul == 2
    assert irreps[1].ir.j == 1 / 2

    assert irreps[2].mul == 1
    assert irreps[2].ir.j == 1

    assert irreps.dim == 1 + 2 * 2 + 3

    irreps = cue.Irreps("SU2", [(3, 0), (2, 1 / 2), (1, 1)])
    assert len(irreps) == 3
    assert irreps[1].ir == cue.SU2(1 / 2)


def test_o3():
    irreps = cue.Irreps("O3", "0e + 1o + 2e")
    assert len(irreps) == 3

    assert irreps[0].mul == 1
    assert irreps[0].ir.l == 0
    assert irreps[0].ir.p == 1

    assert irreps[1].mul == 1
    assert irreps[1].ir.l == 1
    assert irreps[1].ir.p == -1

    assert irreps[2].mul == 1
    assert irreps[2].ir.l == 2
    assert irreps[2].ir.p == 1

    assert irreps.dim == 1 + 3 + 5


def test_so3():
    irreps = cue.Irreps("SO3", "0 + 1 + 2")
    assert len(irreps) == 3

    assert irreps[0].mul == 1
    assert irreps[0].ir.l == 0

    assert irreps[1].mul == 1
    assert irreps[1].ir.l == 1

    assert irreps[2].mul == 1
    assert irreps[2].ir.l == 2

    assert irreps.dim == 1 + 3 + 5


def test_iter():
    irreps = cue.Irreps("SU2", "1 + 3x2")

    ((mul1, rep1), (mul2, rep2)) = irreps

    assert mul1 == 1
    assert rep1.j == 1
    assert mul2 == 3
    assert rep2.j == 2


def test_getitem():
    irreps = cue.Irreps("SU2", "1/2 + 16x2")

    assert irreps[0].mul == 1
    assert irreps[0].ir.j == 1 / 2

    assert irreps[1].mul == 16
    assert irreps[1].ir.j == 2


def test_contains():
    irreps = cue.Irreps("SU2", "1/2 + 16x2")

    assert 1 / 2 in irreps
    assert 3 / 2 not in irreps
    assert cue.SU2(2) in irreps


def test_repr():
    irreps = cue.Irreps("O3", "0e + 1o + 3x2e")

    assert repr(irreps) == "0e+1o+3x2e"


def test_count():
    irreps = cue.Irreps("SU2", "1/2 + 16x2 + 1/2")

    assert irreps.count(1 / 2) == 2
    assert irreps.count(2) == 16
    assert irreps.count(cue.SU2(2)) == 16
    assert irreps.count(12) == 0


def test_dim():
    irreps = cue.Irreps("SU2", "1/2 + 16x2")
    assert irreps.dim == 2 + 16 * 5


def test_num_irreps():
    irreps = cue.Irreps("SU2", "1/2 + 16x2")
    assert irreps.num_irreps == 17


def test_slices():
    irreps = cue.Irreps("SU2", "1/2 + 16x2")
    assert irreps.slices() == [slice(0, 2), slice(2, 2 + 16 * 5)]


def test_is_scalar():
    assert cue.Irreps("O3", "0e + 1o").is_scalar() is False
    assert cue.Irreps("O3", "0e + 0x1o").is_scalar() is False
    assert cue.Irreps("O3", "0e + 0e").is_scalar() is True


def test_add():
    assert cue.Irreps("SU2", "1/2") + cue.Irreps("SU2", "1 + 2") == cue.Irreps(
        "SU2", "1/2 + 1 + 2"
    )
    assert "1/2" + cue.Irreps("SU2", "1") == cue.Irreps("SU2", "1/2 + 1")


def test_mul():
    assert 3 * cue.Irreps("SU2", "1/2") == cue.Irreps("SU2", "3x1/2")


def test_floordiv():
    assert cue.Irreps("SU2", "3x1/2") // 3 == cue.Irreps("SU2", "1/2")


def test_eq():
    assert cue.Irreps("SU2", "1/2") == "1/2"
    assert not (cue.Irreps("SU2", "1/2") == cue.Irreps("SU2", "1"))
    assert not (cue.Irreps("SU2", "1/2 + 0x2") == cue.Irreps("SU2", "1/2"))
    assert not (cue.Irreps("SU2", "1/2") == "sapristi !")


def test_merge_consecutive():
    assert cue.Irreps("SU2", "1+1+0x2").merge_consecutive() == cue.Irreps(
        "SU2", "2x1+0x2"
    )


def test_remove_zero_multiplicities():
    assert cue.Irreps("SU2", "1+1+0x2").remove_zero_multiplicities() == cue.Irreps(
        "SU2", "1+1"
    )


def test_simplify():
    assert cue.Irreps("SU2", "1+1+0x2+3+1").simplify() == cue.Irreps("SU2", "2x1+3+1")


def test_sort():
    irreps, perm, inv = cue.Irreps("SU2", "1+1+0x2+3+1").sort()

    assert irreps == cue.Irreps("SU2", "1+1+1+0x2+3")
    assert perm == (0, 1, 3, 4, 2)
    assert inv == (0, 1, 4, 2, 3)


def test_regroup():
    assert cue.Irreps("SU2", "1+1+0x2+3+1").regroup() == cue.Irreps("SU2", "3x1+3")


def test_set_mul():
    assert cue.Irreps("SU2", "1+1+0x2+3+1").set_mul(2) == "2x1+2x1+2x2+2x3+2x1"


def test_filter():
    assert cue.Irreps("SU2", "0 + 1/2 + 1 + 3x3/2").filter(
        keep=lambda mul_ir: mul_ir.ir.j > 1
    ) == cue.Irreps("SU2", "3x3/2")
    assert cue.Irreps("SU2", "0 + 1/2 + 1 + 3x3/2").filter(
        drop=lambda mul_ir: mul_ir.ir.j > 1
    ) == cue.Irreps("SU2", "0+1/2+1")
    assert cue.Irreps("SU2", "0 + 1/2 + 1 + 3x3/2").filter(drop=1) == cue.Irreps(
        "SU2", "0+1/2+3x3/2"
    )
    assert cue.Irreps("SU2", "0 + 1/2 + 1 + 3x3/2").filter(
        keep="2x1/2 + 1"
    ) == cue.Irreps("SU2", "1/2+1")


def test_init():
    irreps_su2 = cue.Irreps("SU2", "1 + 2")
    with pytest.raises(ValueError):
        cue.Irreps("SO3", irreps_su2)

    with pytest.raises(ValueError):
        cue.Irreps("SO3", [(12, "1", -1)])

    with pytest.raises(ValueError):
        cue.Irreps("SO3", [(12, "1/2")])



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools

import numpy as np
import pytest

import cuequivariance as cue


@pytest.mark.parametrize("Group", [cue.SU2, cue.SO3, cue.O3])
def test_clebsch_gordan(Group):
    it = Group.iterator()
    irreps = [next(it) for _ in range(4)]

    for r1, r2, r3 in itertools.combinations_with_replacement(irreps, 3):
        C = Group.clebsch_gordan(r1, r2, r3)

        a1 = np.einsum("zijk,giu->gzujk", C, r1.X)
        a2 = np.einsum("zijk,gju->gziuk", C, r2.X)
        a3 = np.einsum("zijk,guk->gziju", C, r3.X)  # Note the transpose

        np.testing.assert_allclose(a1 + a2, a3, atol=1e-10, rtol=0)



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import warnings
from typing import Any, Generator, Optional, Union

import cuequivariance as cue


def default_layout(layout: Optional[cue.IrrepsLayout]) -> cue.IrrepsLayout:
    if layout is None:
        warnings.warn(
            "layout is not specified, defaulting to cue.mul_ir. This is the layout used in the e3nn library."
            " We use it as the default layout for compatibility with e3nn."
            " However, the cue.ir_mul layout is faster and more memory efficient."
            " Please specify the layout explicitly to avoid this warning."
        )
        return cue.mul_ir
    if isinstance(layout, str):
        return cue.IrrepsLayout[layout]
    return layout


def assert_same_group(*irreps_: cue.Irreps) -> None:
    group = irreps_[0].irrep_class
    for irreps in irreps_[1:]:
        if group != irreps.irrep_class:
            raise ValueError("The provided irreps are not of the same group.")


def default_irreps(
    *irreps_: Union[cue.Irreps, Any],
) -> Generator[cue.Irreps, None, None]:
    for irreps in irreps_:
        if isinstance(irreps, cue.Irreps):
            yield irreps
        else:
            warnings.warn(
                "irreps should be of type cue.Irreps, converting to cue.Irreps(cue.O3, ...) for compatibility with e3nn."
            )
            yield cue.Irreps(cue.O3, irreps)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses
import itertools
from collections import defaultdict

IVARS = "abcdefghijklmnopqrstuvwxyz"
OVARS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"


@dataclasses.dataclass(init=False, frozen=True)
class Operation:
    """Descriptor mapping input/output buffers to tensor product operands.

    The buffers are identified by their index (0, 1, 2, ...).
    The order of the buffers corresponds to the order of the operands.

    Example:

        This list of operations would typically be used for the symmetric contraction operation.

        >>> ops = [
        ...     Operation((0, 1, 2)),
        ...     Operation((0, 1, 1, 2)),
        ...     Operation((0, 1, 1, 1, 2)),
        ... ]
        >>> print(Operation.list_to_string(ops, 2, 1))
        (a, b) -> (C)
          a b C
          a b b C
          a b b b C
    """

    buffers: tuple[int, ...]

    def __init__(self, buffers: tuple[int, ...] | Operation):
        if isinstance(buffers, Operation):
            buffers = buffers.buffers
        assert len(buffers) > 0, buffers
        assert all(isinstance(b, int) for b in buffers), buffers
        assert all(i >= 0 for i in buffers), buffers
        object.__setattr__(self, "buffers", tuple(int(b) for b in buffers))

    def __repr__(self):
        return f"Operation({self.buffers})"

    def to_letters(self, num_inputs: int) -> list[str]:
        return [IVARS[b] if b < num_inputs else OVARS[b] for b in self.buffers]

    @staticmethod
    def list_to_string(
        operations: list[Operation], num_inputs: int, num_outputs: int
    ) -> str:
        i = ", ".join(IVARS[:num_inputs])
        o = ", ".join(OVARS[num_inputs : num_inputs + num_outputs])
        s = f"({i}) -> ({o})"
        for op in operations:
            s += "\n  " + " ".join(op.to_letters(num_inputs))
        return s

    def __lt__(self, value):
        assert isinstance(value, Operation)
        return (len(self.buffers), self.buffers) < (len(value.buffers), value.buffers)

    def __hash__(self) -> int:
        return hash(self.buffers)

    def __eq__(self, value):
        assert isinstance(value, Operation)
        return self.buffers == value.buffers

    def permute_operands(self, permutation: tuple[int, ...]) -> Operation:
        return Operation(tuple(self.buffers[p] for p in permutation))

    def move_operand_last(self, operand: int) -> Operation:
        buffers = list(self.buffers)
        b = buffers.pop(operand)
        buffers.append(b)
        return Operation(tuple(buffers))

    def input_operands_buffers(self, num_inputs: int) -> list[tuple[int, int]]:
        return [(op, i) for op, i in enumerate(self.buffers) if i < num_inputs]

    def output_operand_buffer(self, num_inputs: int) -> tuple[int, int]:
        def _raise():
            raise ValueError(
                f"Operation must have exactly one output buffer. {self=} {num_inputs=}"
            )

        result = None
        for op, i in enumerate(self.buffers):
            if i < num_inputs:
                continue
            if result is not None:
                _raise()
            result = (op, i)
        if result is None:
            _raise()

        return result

    def input_buffers(self, num_inputs: int) -> list[int]:
        return [i for i in self.buffers if i < num_inputs]

    def output_buffer(self, num_inputs: int) -> int:
        return self.output_operand_buffer(num_inputs)[1]

    def transpose(
        self,
        is_undefined_primal: list[bool],
        has_cotangent: list[bool],
    ) -> Operation | None:
        """
        Args:
            is_undefined_primal (list[bool]): whether the primal is undefined
            has_cotangent (list[bool]): whether the cotangent is defined

        Returns:
            Operation: the transposed operation, if any
                in the returned operation, the buffers are:
                 - new inputs: defined primals + cotangents (=True)
                 - new outputs: undefined primals
        """
        # number of input buffers in the original operation
        # note that self might not involve all input buffers
        num_inputs = len(is_undefined_primal)

        if not has_cotangent[self.output_buffer(num_inputs) - num_inputs]:
            # The output buffer of self has no cotangent,
            # so there is no contribution to the primal.
            return None

        num_undef_primal = 0
        for i in self.input_buffers(num_inputs):
            if is_undefined_primal[i]:
                num_undef_primal += 1

        if num_undef_primal > 1:
            raise ValueError(
                f"Operation must have at most one undefined primal input. {self=} {is_undefined_primal=}."
                " Otherwise it means that the operation has not been linearized correctly."
            )
        if num_undef_primal == 0:
            # The operation has no undefined primal as input
            return None

        new_num_inputs = 0
        new_num_outputs = 0
        primals_to_new_input: list[int | None] = []
        primals_to_new_output: list[int | None] = []

        for undef in is_undefined_primal:
            if undef:
                primals_to_new_input.append(None)
                primals_to_new_output.append(new_num_outputs)
                new_num_outputs += 1
            else:
                primals_to_new_input.append(new_num_inputs)
                primals_to_new_output.append(None)
                new_num_inputs += 1

        outputs_to_new_input: list[int | None] = []
        for has in has_cotangent:
            if has:
                outputs_to_new_input.append(new_num_inputs)
                new_num_inputs += 1
            else:
                outputs_to_new_input.append(None)

        # output buffers are identified by their index being >= new_num_inputs
        primals_to_new_output = [
            None if i is None else new_num_inputs + i for i in primals_to_new_output
        ]

        new_buffers = []
        for i in self.buffers:
            if i < num_inputs:
                if is_undefined_primal[i]:
                    new_buffers.append(primals_to_new_output[i])
                else:
                    new_buffers.append(primals_to_new_input[i])
            else:
                new_buffers.append(outputs_to_new_input[i - num_inputs])

        return Operation(tuple(new_buffers))

    def jvp(self, has_tangent: list[bool]) -> list[Operation]:
        """
        Args:
            has_tangent (list[bool]): whether the input has a tangent

        Returns:
            list[Operation]: the JVPs of the operation
                in the returned operations, the buffers are:
                 - new inputs: original inputs + tangents (=True)
                 - new outputs: original outputs
        """
        # number of input buffers in the original operation
        # note that self might not involve all input buffers
        num_inputs = len(has_tangent)

        new_num_inputs = num_inputs
        mapping: list[int | None] = []
        for has in has_tangent:
            if has:
                mapping.append(new_num_inputs)
                new_num_inputs += 1
            else:
                mapping.append(None)

        jvps = []
        for op, i in self.input_operands_buffers(num_inputs):
            if has_tangent[i]:
                new_buffers = list(self.buffers)
                new_buffers[op] = mapping[i]
                op, i = self.output_operand_buffer(num_inputs)
                new_buffers[op] += new_num_inputs - num_inputs
                jvps.append(Operation(new_buffers))

        return jvps

    def operands_with_identical_buffers(self) -> frozenset[frozenset[int]]:
        """
        Groups of operands sharing the same buffer.
        """
        bid_to_oid = defaultdict(list)
        for oid, b in enumerate(self.buffers):
            bid_to_oid[b].append(oid)
        return frozenset(map(frozenset, bid_to_oid.values()))

    @staticmethod
    def group_by_idential_buffers(
        operations: list[Operation],
    ) -> list[tuple[frozenset[frozenset[int]], list[Operation]]]:
        """
        Args:
            operations (list[Operation]): the operations to group
            num_inputs (int): the number of input buffers

        Returns:
            list of tuples: Each tuple contains:
                - frozenset of frozensets of operands bound to identical buffers
                - list of operations
        """
        return [
            (p, list(group))
            for p, group in itertools.groupby(
                operations, key=Operation.operands_with_identical_buffers
            )
        ]

    @staticmethod
    def group_by_operational_symmetries(
        symmetries: list[tuple[int, ...]],
        operations: list[Operation],
    ) -> list[tuple[int, Operation]]:
        """
        Args:
            symmetries (list[tuple[int, ...]]): the permutation group
            operations (list[Operation]): the operations to group

        Returns:
            list of tuples: Each tuple contains:
                - multiplicity (int)
                - a representative operation
        """

        def partition(operation: Operation) -> tuple[int, ...]:
            return frozenset(operation.permute_operands(perm) for perm in symmetries)

        groups = []
        for _, group in itertools.groupby(operations, key=partition):
            group = sorted(group)
            groups.append((len(group), group[0]))
        return groups



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import cuequivariance as cue


def transpose(
    irreps: cue.Irreps, source: cue.IrrepsLayout, target: cue.IrrepsLayout
) -> cue.EquivariantPolynomial:
    """Transpose the irreps layout of a tensor."""
    d = cue.SegmentedTensorProduct(
        operands_and_subscripts=[
            (cue.SegmentedOperand(ndim=2), "ui" if source == cue.mul_ir else "iu"),
            (cue.SegmentedOperand(ndim=2), "ui" if target == cue.mul_ir else "iu"),
        ]
    )
    for mul, ir in irreps:
        d.add_path(None, None, c=1, dims={"u": mul, "i": ir.dim})
    return cue.EquivariantPolynomial(
        [cue.IrrepsAndLayout(irreps, source)],
        [cue.IrrepsAndLayout(irreps, target)],
        cue.SegmentedPolynomial.eval_last_operand(d),
    )



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
import math
from typing import Generator, Tuple

# we cannot import cuequivariance as cue because of circular import
from cuequivariance.segmented_polynomials import SegmentedTensorProduct, Subscripts


def dispatch(
    descriptor: SegmentedTensorProduct,
    targets: list[Subscripts],
    permutation_mode: str,
) -> Generator[Tuple[SegmentedTensorProduct, Tuple[int, ...]], None, None]:
    """Dispatch a descriptor to a target subscripts.

    Args:
        descriptor (SegmentedTensorProduct): The descriptor to dispatch.
        targets (list of stp.Subscripts): A list of target subscripts.
        permutation_mode (str): The permutation mode. One of "permute_none", "permute_all", "permute_all_but_last".

    Yields:
        Tuple[stp.SegmentedTensorProduct, Tuple[int, ...]]:
            A tuple of the dispatched descriptor and the permutation of the operands.

    Note:
        The function tries to dispatch the descriptor to the target subscripts. If the descriptor
        is not dispatchable to the target subscripts, it will try to flatten the descriptor progressively
        and dispatch the flattened descriptor to the target subscripts. The function will yield all the
        possible dispatches found.
    """
    targets = [Subscripts(subscripts) for subscripts in targets]
    targets = [
        subscripts
        for subscripts in targets
        if subscripts.num_operands == descriptor.num_operands
    ]

    if permutation_mode == "permute_none":
        permutations = [tuple(range(descriptor.num_operands))]
    elif permutation_mode == "permute_all":
        permutations = list(itertools.permutations(range(descriptor.num_operands)))
    elif permutation_mode == "permute_all_but_last":
        permutations = [
            p + (descriptor.num_operands - 1,)
            for p in itertools.permutations(range(descriptor.num_operands - 1))
        ]
    else:
        raise ValueError(f"unknown permutation_mode: {permutation_mode}")

    # Squeeze all the channels of extent 1
    descriptor = descriptor.squeeze_modes()

    # Consolidate all the repeated channels e.g. ab -> a
    descriptor = descriptor.consolidate_modes()

    for subscripts in targets:
        for perm in permutations:
            d = descriptor.permute_operands(perm)

            for mapping in d.subscripts.is_subset_of(subscripts):
                d = d.add_or_rename_modes(subscripts, mapping=mapping)
                yield d, perm

    # Flatten one channel at a time, starting from the coefficients channel
    flattenable_powerset = descriptor.subscripts.flattenable_powerset()
    dims = descriptor.get_dimensions_dict()
    flattenable_powerset = sorted(
        flattenable_powerset,
        key=lambda channels: (
            len(channels),
            sum(-1 for ch in channels if ch in descriptor.subscripts.coefficients),
            math.prod(max(dims[ch]) for ch in channels),
        ),
    )

    for channels in flattenable_powerset:
        d = descriptor.flatten_modes(channels).remove_zero_paths()
        yield from dispatch(d, targets, permutation_mode)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .subscripts import Subscripts
from .path import Path
from .segmented_operand import SegmentedOperand
from .segmented_tensor_product import SegmentedTensorProduct
from .dot import dot, trace

from .evaluate import compute_last_operand, primitive_compute_last_operand
from .dispatch import dispatch

from .operation import Operation
from .segmented_polynomial import SegmentedPolynomial


__all__ = [
    "Subscripts",
    "Path",
    "SegmentedOperand",
    "SegmentedTensorProduct",
    "dot",
    "trace",
    "compute_last_operand",
    "primitive_compute_last_operand",
    "dispatch",
    "Operation",
    "SegmentedPolynomial",
]



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, List, Optional, Sequence, Tuple, Union

import numpy as np

import cuequivariance as cue
from cuequivariance.group_theory.irreps_array.irrep_utils import into_list_of_irrep


# This class is inspired by https://github.com/e3nn/e3nn-jax/blob/245e17eb23deaccad9f2c9cfd40fe40515e3c074/e3nn_jax/_src/irreps_array.py
@dataclass(frozen=True, init=False)
class NumpyIrrepsArray:
    r"""
    Data Array transforming according to the irreducible representations of a Lie group.

    Args:
        irreps (Irreps or str): Irreps of the data array.
        array (np.ndarray): Data array.
        layout (IrrepsLayout, optional): Memory layout of the data array.
    """

    irreps: cue.Irreps
    array: np.ndarray
    layout: cue.IrrepsLayout

    def __init__(
        self,
        irreps: Union[cue.Irreps, str],
        array: np.ndarray,
        layout: Optional[cue.IrrepsLayout] = None,
    ):
        irreps = cue.Irreps(irreps)
        layout = cue.IrrepsLayout.as_layout(layout)

        if not hasattr(array, "shape"):
            raise TypeError(f"Expected array with shape, got {type(array)}")

        if array.shape[-1] != irreps.dim:
            raise ValueError(
                f"Last dimension of array should be {irreps.dim}, got {array.shape[-1]}"
            )

        object.__setattr__(self, "irreps", irreps)
        object.__setattr__(self, "array", array)
        object.__setattr__(self, "layout", layout)

    @property
    def shape(self) -> Tuple[int, ...]:
        """Shape of the data array."""
        return self.array.shape

    @property
    def dtype(self):
        """Data type of the data array."""
        return self.array.dtype

    @property
    def ndim(self) -> int:
        """Number of dimensions of the data array."""
        return self.array.ndim

    def __len__(self) -> int:
        return len(self.array)

    def __repr__(self) -> str:
        r = str(self.array)
        if "\n" in r:
            return f"{self.irreps}\n{r}"
        return f"{self.irreps} {r}"

    @property
    def segments(self) -> List[np.ndarray]:
        """List of segments of the data array."""
        leading_shape = self.array.shape[:-1]
        return [
            np.reshape(self.array[..., i], leading_shape + self.layout.shape(mul_ir))
            for i, mul_ir in zip(self.irreps.slices(), self.irreps)
        ]

    def __eq__(self: NumpyIrrepsArray, other: NumpyIrrepsArray) -> bool:
        if not isinstance(other, NumpyIrrepsArray):
            return NotImplemented
        return (
            self.irreps == other.irreps
            and np.array_equal(self.array, other.array)
            and self.layout == other.layout
        )

    def __neg__(self: NumpyIrrepsArray) -> NumpyIrrepsArray:
        return NumpyIrrepsArray(self.irreps, -self.array, self.layout)

    def reshape(self, shape: Tuple[int, ...]) -> NumpyIrrepsArray:
        if not (shape[-1] == -1 or shape[-1] == self.irreps.dim):
            raise ValueError(
                f"Last dimension of shape should be -1 or {self.irreps.dim}, got {shape[-1]}"
            )
        return NumpyIrrepsArray(self.irreps, self.array.reshape(shape), self.layout)

    def simplify(self) -> NumpyIrrepsArray:
        if self.layout == cue.mul_ir:
            return NumpyIrrepsArray(self.irreps.simplify(), self.array, self.layout)

        return self.change_layout(cue.mul_ir).simplify().change_layout(self.layout)

    def merge_consecutive(self) -> NumpyIrrepsArray:
        return NumpyIrrepsArray(
            self.irreps.merge_consecutive(), self.array, self.layout
        )

    def sort(self) -> NumpyIrrepsArray:
        s = self.irreps.sort()
        # s.irreps, s.perm, s.inv
        segments = self.segments
        sorted_segments = [segments[i] for i in s.inv]
        return from_segments(s.irreps, sorted_segments, self.layout, self.shape[:-1])

    def regroup(self) -> NumpyIrrepsArray:
        return self.sort().simplify()

    def filter(
        self,
        *,
        keep: Union[str, Sequence[cue.Irrep], Callable[[cue.MulIrrep], bool]] = None,
        drop: Union[str, Sequence[cue.Irrep], Callable[[cue.MulIrrep], bool]] = None,
    ) -> NumpyIrrepsArray:
        if keep is not None:
            if drop is not None:
                raise ValueError("Only one of `keep` or `drop` must be defined.")
            else:
                return self._filter_keep(keep)
        else:
            if drop is not None:
                return self._filter_drop(drop)
            else:
                return self

    def _filter_keep(
        self, keep: Union[str, Sequence[cue.Irrep], Callable[[cue.MulIrrep], bool]]
    ):
        if callable(keep):
            return from_segments(
                self.irreps.filter(keep=keep),
                [
                    segment
                    for segment, mulrep in zip(self.segments, self.irreps)
                    if keep(mulrep)
                ],
                self.layout,
                self.shape[:-1],
            )

        keep = into_list_of_irrep(self.irreps.irrep_class, keep)

        if not all(isinstance(rep, cue.Irrep) for rep in keep):
            raise ValueError(f"Invalid `keep` {keep}, expected a list of Irrep")

        return from_segments(
            self.irreps.filter(keep=keep),
            [
                segment
                for segment, mulrep in zip(self.segments, self.irreps)
                if mulrep.ir in keep
            ],
            self.layout,
            self.shape[:-1],
        )

    def _filter_drop(
        self, drop: Union[str, Sequence[cue.Irrep], Callable[[cue.MulIrrep], bool]]
    ):
        if callable(drop):
            return from_segments(
                self.irreps.filter(drop=drop),
                [
                    segments
                    for segments, mulrep in zip(self.segments, self.irreps)
                    if not drop(mulrep)
                ],
                self.layout,
                self.shape[:-1],
            )

        drop = into_list_of_irrep(self.irreps.irrep_class, drop)

        if not all(isinstance(rep, cue.Irrep) for rep in drop):
            raise ValueError(f"Invalid `drop` {drop}, expected a list of Irrep")

        return from_segments(
            self.irreps.filter(drop=drop),
            [
                segments
                for segments, mulrep in zip(self.segments, self.irreps)
                if mulrep.ir not in drop
            ],
            self.layout,
            self.shape[:-1],
        )

    def change_layout(self, layout: cue.IrrepsLayout) -> NumpyIrrepsArray:
        if layout == self.layout:
            return self
        segments = [np.moveaxis(x, -1, -2) for x in self.segments]
        return from_segments(self.irreps, segments, layout, self.shape[:-1])


def from_segments(
    irreps: cue.Irreps,
    segments: List[np.ndarray],
    layout: cue.IrrepsLayout,
    leading_shape: Tuple[int, ...],
) -> NumpyIrrepsArray:
    if len(segments) != len(irreps):
        raise ValueError(f"Expected {len(irreps)} segments, got {len(segments)}")

    for segment, mul_ir in zip(segments, irreps):
        if segment.shape[:-2] != leading_shape:
            raise ValueError(
                f"Expected shape[:-2] {leading_shape}, got {segment.shape[:-2]}"
            )
        if segment.shape[-2:] != layout.shape(mul_ir):
            raise ValueError(
                f"Expected shape[-2:] {layout.shape(mul_ir)}, got {segment.shape[-2:]}"
            )

    if len(segments) > 0:
        array = np.concatenate(
            [
                np.reshape(segment, segment.shape[:-2] + (mul * ir.dim,))
                for segment, (mul, ir) in zip(segments, irreps)
            ],
            axis=-1,
        )
    else:
        array = np.empty(leading_shape + (0,))
    return NumpyIrrepsArray(irreps, array, layout)


def concatenate(
    arrays: Union[
        list[cue.Irreps],
        list[Union[cue.IrrepsAndLayout]],
        list[NumpyIrrepsArray],
    ],
) -> Union[cue.Irreps, cue.IrrepsAndLayout, NumpyIrrepsArray]:
    if len(arrays) == 0:
        raise ValueError("Expected at least one input")

    if all(isinstance(array, cue.IrrepsAndLayout) for array in arrays):
        assert len({x.layout for x in arrays}) == 1
        return cue.IrrepsAndLayout(
            concatenate([x.irreps for x in arrays]), arrays[0].layout
        )

    if all(isinstance(array, cue.Irreps) for array in arrays):
        return sum(arrays, cue.Irreps(arrays[0].irrep_class, []))

    if not all(isinstance(array, NumpyIrrepsArray) for array in arrays):
        raise TypeError("Expected a list of IrrepsArray")

    if not all(
        array.irreps.irrep_class == arrays[0].irreps.irrep_class for array in arrays
    ):
        raise ValueError("Expected all arrays to have the same irrep class")

    if not all(array.layout == arrays[0].layout for array in arrays):
        raise ValueError("Expected all arrays to have the same layout")

    irreps = cue.Irreps(
        arrays[0].irreps.irrep_class,
        sum([[mul_ir for mul_ir in array.irreps] for array in arrays], []),
    )

    return NumpyIrrepsArray(
        irreps,
        np.concatenate([array.array for array in arrays], axis=-1),
        arrays[0].layout,
    )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import cuequivariance as cue


def test_operation_transpose():
    ope = cue.Operation([0, 1, 2, 3])
    assert ope.transpose([True, False, False], [True]) == cue.Operation([3, 0, 1, 2])

    ope = cue.Operation([0, 1, 2, 4])
    assert ope.transpose([True, False, False, True], [True]) == cue.Operation(
        [3, 0, 1, 2]
    )

    ope = cue.Operation([4, 2, 3, 1])
    assert ope.transpose([True, False, False, True], [True]) == cue.Operation(
        [2, 1, 4, 0]
    )


def test_operation_cube():
    ope = cue.Operation([0, 0, 0, 1])  # x^3
    jvps = ope.jvp([True])
    assert jvps == [
        cue.Operation((1, 0, 0, 2)),
        cue.Operation((0, 1, 0, 2)),
        cue.Operation((0, 0, 1, 2)),
    ]

    [(mul, ope)] = cue.Operation.group_by_operational_symmetries(
        [
            (0, 1, 2, 3),
            (0, 2, 1, 3),
            (1, 0, 2, 3),
            (1, 2, 0, 3),
            (2, 0, 1, 3),
            (2, 1, 0, 3),
        ],
        jvps,
    )
    assert mul == 3
    assert ope == cue.Operation((0, 0, 1, 2))  # d/dx (x^3) = 3x^2 dx

    ope = ope.transpose([False, True], [True])
    assert ope == cue.Operation((0, 0, 2, 1))

    [(groupped_operands, [ope])] = cue.Operation.group_by_idential_buffers([ope])
    assert groupped_operands == frozenset(
        {frozenset({3}), frozenset({0, 1}), frozenset({2})}
    )
    assert ope == cue.Operation((0, 0, 2, 1))



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import copy
import dataclasses
import warnings
from typing import Optional, Sequence, Union

import cuequivariance as cue


@dataclasses.dataclass(init=False, frozen=True)
class EquivariantTensorProduct:
    """
    Descriptor of an equivariant tensor product.
    This class is a wrapper around a list of :class:`STP <cuequivariance.SegmentedTensorProduct>`.
    While an STP is a single homogeneous polynomial without specification of the role of each operand,
    an ETP determines the role of each operand (input or output), the representation of each operand (irreps),
    and the layout of each operand (multiplicity first or irreducible representation first).

    Requirements:
        - An ETP must contain at least one :class:`STP <cuequivariance.SegmentedTensorProduct>`.
        - Each STP must have at least one operand (the output).

    Examples:
        +------+--------+--------+--------+--------+---------------------------------------------------+
        |      | Input0 | Input1 | Input2 | Output |                    Comment                        |
        +======+========+========+========+========+===================================================+
        | STP0 |   x    |   x    |   x    |   x    |  common case, the number of operands is the same  |
        +------+--------+--------+--------+--------+---------------------------------------------------+
        | STP1 |   x    |   x    |        |   x    |  some inputs are not used by all STPs             |
        +------+--------+--------+--------+--------+---------------------------------------------------+
        | STP2 |   x    |        |        |   x    |  -- " --                                          |
        +------+--------+--------+--------+--------+---------------------------------------------------+
        | STP3 |        |        |        |   x    |  -- " --                                          |
        +------+--------+--------+--------+--------+---------------------------------------------------+
        | STP4 |   x    |   x    |  x x x |   x    |  the last input is fed multiple times             |
        +------+--------+--------+--------+--------+---------------------------------------------------+

    .. rubric:: Methods
    """

    operands: tuple[cue.Rep, ...]
    ds: list[cue.SegmentedTensorProduct]

    def __init__(
        self,
        d: Union[cue.SegmentedTensorProduct, Sequence[cue.SegmentedTensorProduct]],
        operands: list[cue.Rep],
        symmetrize: bool = True,
    ):
        warnings.warn(
            "EquivariantTensorProduct is deprecated and will be removed in a future version. "
            "Please use EquivariantPolynomial instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        operands = tuple(operands)
        if isinstance(d, cue.SegmentedTensorProduct):
            assert len(operands) == d.num_operands
            for oid in range(d.num_operands):
                assert operands[oid].dim == d.operands[oid].size
            ds = [d]
        else:
            ds = []

            nin = len(operands) - 1

            for d in list(d):
                # all non-repeated input operands are the same
                for i in range(nin - 1):
                    if not (i < d.num_operands - 1):
                        continue

                    if operands[i].dim != d.operands[i].size:
                        raise ValueError(
                            f"Input {i} size mismatch: {operands[i]} vs {d.operands[i]}"
                        )
                # the repeated input operand is the same
                assert len(operands) >= 2
                for d_ope in d.operands[nin - 1 : -1]:
                    if operands[-2].dim != d_ope.size:
                        raise ValueError(
                            f"Last input size mismatch: {operands[-2]} vs {d_ope}"
                        )
                if operands[-1].dim != d.operands[-1].size:
                    raise ValueError(
                        f"Output size mismatch: {operands[-1]} vs {d.operands[-1]}"
                    )

                if symmetrize:
                    d = d.symmetrize_operands(range(nin - 1, d.num_operands - 1))
                ds.append(d)

            # all non-repeated inputs have the same operands
            for i in range(nin - 1):
                assert len({d.operands[i] for d in ds if i < d.num_operands - 1}) <= 1
            # all the repeated inputs have the same operands
            if len({o for d in ds for o in d.operands[nin - 1 : -1]}) > 1:
                tmp = {o for d in ds for o in d.operands[nin - 1 : -1]}
                raise ValueError(f"Different operands for the last input, {tmp}")
            # all the output operands are the same
            assert len({d.operands[-1] for d in ds}) == 1

        object.__setattr__(self, "operands", operands)
        object.__setattr__(self, "ds", ds)

    def __hash__(self) -> int:
        return hash((self.operands, tuple(self.ds)))

    def __mul__(self, factor: float) -> EquivariantTensorProduct:
        return EquivariantTensorProduct([d * factor for d in self.ds], self.operands)

    def __rmul__(self, factor: float) -> EquivariantTensorProduct:
        return self.__mul__(factor)

    @property
    def d(self) -> cue.SegmentedTensorProduct:
        assert len(self.ds) == 1
        return self.ds[0]

    @property
    def num_operands(self) -> int:
        return len(self.operands)

    @property
    def num_inputs(self) -> int:
        return self.num_operands - 1

    @property
    def inputs(self) -> tuple[cue.Rep, ...]:
        return self.operands[:-1]

    @property
    def output(self) -> cue.Rep:
        return self.operands[-1]

    def _degrees(self, i: int) -> set[int]:
        assert 0 <= i < self.num_inputs
        deg = set()
        if i == self.num_inputs - 1:
            for d in self.ds:
                deg.add(max(0, d.num_operands - self.num_inputs))
        else:
            for d in self.ds:
                deg.add(1 if i < d.num_operands - 1 else 0)
        return deg

    def __repr__(self) -> str:
        irs = [str(ope.irreps) for ope in self.operands]
        inputs = irs[:-1]
        output = irs[-1]

        for i in range(self.num_inputs):
            d = self._degrees(i)
            if d == set():
                inputs[i] = f"({inputs[i]})"
            elif len(d) > 1:
                inputs[i] = f"({inputs[i]})^({min(d)}..{max(d)})"
            elif d != {1}:
                inputs[i] = f"({inputs[i]})^{min(d)}"

        return f"{self.__class__.__name__}({' x '.join(inputs)} -> {output})"

    def permute_operands(
        self, permutation: tuple[int, ...]
    ) -> EquivariantTensorProduct:
        """Permute the operands of the tensor product."""
        assert sorted(permutation) == list(range(self.num_operands))
        assert all(d.num_operands == self.num_operands for d in self.ds)
        return EquivariantTensorProduct(
            [d.permute_operands(permutation) for d in self.ds],
            tuple(self.operands[pid] for pid in permutation),
        )

    def move_operand(self, src: int, dst: int) -> EquivariantTensorProduct:
        """Move an operand to a new position."""
        if src < 0:
            src += self.num_operands
        if dst < 0:
            dst += self.num_operands
        return self.permute_operands(
            tuple(
                {src: dst, dst: src}.get(oid, oid) for oid in range(self.num_operands)
            )
        )

    def move_operand_first(self, src: int) -> EquivariantTensorProduct:
        """Move an operand to the front."""
        return self.move_operand(src, 0)

    def move_operand_last(self, src: int) -> EquivariantTensorProduct:
        """Move an operand to the back."""
        return self.move_operand(src, -1)

    def squeeze_modes(self, modes: Optional[str] = None) -> EquivariantTensorProduct:
        """Squeeze the modes."""
        return EquivariantTensorProduct(
            [d.squeeze_modes(modes) for d in self.ds], self.operands
        )

    def consolidate_paths(self) -> EquivariantTensorProduct:
        """Consolidate the paths."""
        return EquivariantTensorProduct(
            [d.consolidate_paths() for d in self.ds], self.operands
        )

    def canonicalize_subscripts(self) -> EquivariantTensorProduct:
        """Canonicalize the subscripts."""
        return EquivariantTensorProduct(
            [d.canonicalize_subscripts() for d in self.ds], self.operands
        )

    def flatten_modes(
        self, modes: str, *, skip_zeros: bool = True, force: bool = False
    ) -> EquivariantTensorProduct:
        """Flatten modes."""
        return EquivariantTensorProduct(
            [
                d.flatten_modes(modes, skip_zeros=skip_zeros, force=force)
                for d in self.ds
            ],
            self.operands,
        )

    def all_same_segment_shape(self) -> bool:
        """Whether all the segments have the same shape."""
        return all(d.all_same_segment_shape() for d in self.ds)

    def flatten_coefficient_modes(self) -> EquivariantTensorProduct:
        """Flatten the coefficient modes."""
        return EquivariantTensorProduct(
            [d.flatten_coefficient_modes() for d in self.ds], self.operands
        )

    def map_operands(self, num_operand: int) -> list[int]:
        inputs = list(range(self.num_operands - 1))
        output = self.num_operands - 1

        if num_operand == self.num_operands:
            return inputs + [output]
        if num_operand < self.num_operands:
            return inputs[: num_operand - 1] + [output]
        if num_operand > self.num_operands:
            return inputs + [inputs[-1]] * (num_operand - self.num_operands) + [output]

    def change_layout(
        self, layout: Union[cue.IrrepsLayout, list[cue.IrrepsLayout]]
    ) -> EquivariantTensorProduct:
        if isinstance(layout, Sequence):
            layouts = list(layout)
            assert len(layouts) == self.num_operands
        else:
            layouts = [layout] * self.num_operands
        del layout
        layouts = [cue.IrrepsLayout.as_layout(layout) for layout in layouts]

        def f(d: cue.SegmentedTensorProduct) -> cue.SegmentedTensorProduct:
            ii = self.map_operands(d.num_operands)
            assert len(ii) == d.num_operands

            operands = [self.operands[i] for i in ii]
            layouts_ = [layouts[i] for i in ii]

            new_subscripts = []
            for oid, (operand, layout) in enumerate(zip(operands, layouts_)):
                assert isinstance(operand, cue.IrrepsAndLayout)

                subscripts = d.subscripts.operands[oid]
                if operand.layout == layout:
                    new_subscripts.append(subscripts)
                    continue
                if operand.irreps.layout_insensitive():
                    new_subscripts.append(subscripts)
                    continue
                assert d.operands[oid].num_segments == len(operand.irreps)
                assert len(subscripts) > 0
                if operand.layout == cue.ir_mul:
                    assert [seg[0] for seg in d.operands[oid]] == [
                        ir.dim for _, ir in operand.irreps
                    ]
                    new_subscripts.append(subscripts[1:] + subscripts[:1])
                    continue
                if operand.layout == cue.mul_ir:
                    assert [seg[-1] for seg in d.operands[oid]] == [
                        ir.dim for _, ir in operand.irreps
                    ]
                    new_subscripts.append(subscripts[-1:] + subscripts[:-1])
                    continue
                raise NotImplementedError
            return d.add_or_transpose_modes(
                cue.segmented_polynomials.Subscripts.from_operands(
                    new_subscripts, d.coefficient_subscripts
                )
            )

        return EquivariantTensorProduct(
            [f(d) for d in self.ds],
            [
                cue.IrrepsAndLayout(ope.irreps, layout)
                for ope, layout in zip(self.operands, layouts)
            ],
        )

    def flop_cost(self, batch_size: int) -> int:
        """Compute the number of flops of the tensor product."""
        return sum(d.flop(-1) for d in self.ds) * batch_size

    def memory_cost(
        self, batch_sizes: tuple[int, ...], itemsize: Union[int, tuple[int, ...]]
    ) -> int:
        """Compute the number of memory accesses of the tensor product."""
        assert len(batch_sizes) == self.num_operands
        if isinstance(itemsize, int):
            itemsize = (itemsize,) * self.num_operands
        return sum(
            bs * operand.dim * iz
            for iz, bs, operand in zip(itemsize, batch_sizes, self.operands)
        )

    def backward(self, input: int) -> tuple[EquivariantTensorProduct, tuple[int, ...]]:
        """
        The backward pass of the equivariant tensor product.
        """
        assert input < self.num_inputs

        remove_input = (input < self.num_inputs - 1) or all(
            d.num_operands <= self.num_operands for d in self.ds
        )

        ds = []
        for d in self.ds:
            d = d.move_operand_first(-1)

            if self.num_operands == d.num_operands:
                ds.append(d.move_operand_last(input + 1))
            elif self.num_operands > d.num_operands:
                if input < d.num_operands - 1:
                    ds.append(d.move_operand_last(input + 1))
                else:
                    continue
            else:
                if input < self.num_inputs - 1:
                    ds.append(d.move_operand_last(input + 1))
                else:
                    ds.append(
                        d.move_operand_last(input + 1)
                        * (d.num_operands - self.num_inputs)
                    )

        ii = list(range(self.num_inputs))
        if remove_input:
            oids = [self.num_operands - 1] + ii[:input] + ii[input + 1 :] + [ii[input]]
        else:
            oids = [self.num_operands - 1] + ii + [ii[input]]

        e = EquivariantTensorProduct(ds, tuple(self.operands[i] for i in oids))
        return e, tuple(oids)

    def stp_operand(self, oid: int) -> Optional[cue.SegmentedOperand]:
        # output
        if oid == self.num_operands - 1:
            return self.ds[0].operands[-1]
        # input
        for d in self.ds:
            if oid < d.num_operands:
                return d.operands[oid]
        return None

    @classmethod
    def stack(
        cls, es: Sequence[EquivariantTensorProduct], stacked: list[bool]
    ) -> EquivariantTensorProduct:
        """Stack multiple equivariant tensor products."""
        assert len(es) > 0
        num_operands = es[0].num_operands

        assert all(e.num_operands == num_operands for e in es)
        assert len(stacked) == num_operands

        new_operands = []
        for oid in range(num_operands):
            if stacked[oid]:
                if not all(
                    isinstance(e.operands[oid], cue.IrrepsAndLayout) for e in es
                ):
                    raise NotImplementedError(
                        f"Stacking of {type(es[0].operands[oid])} is not implemented"
                    )
                new_operands.append(cue.concatenate([e.operands[oid] for e in es]))
            else:
                ope = es[0].operands[oid]
                assert all(e.operands[oid] == ope for e in es)
                new_operands.append(ope)

        new_ds: dict[int, cue.SegmentedTensorProduct] = {}
        for eid, e in enumerate(es):
            for d in e.ds:
                d = copy.deepcopy(d)
                ii = e.map_operands(d.num_operands)
                for oid in range(d.num_operands):
                    if stacked[ii[oid]]:
                        for e_ in reversed(es[:eid]):
                            d.insert_segments(oid, 0, e_.stp_operand(ii[oid]).segments)
                        for e_ in es[eid + 1 :]:
                            d.insert_segments(oid, -1, e_.stp_operand(ii[oid]).segments)

                if d.num_operands not in new_ds:
                    new_ds[d.num_operands] = d
                else:
                    for p in d.paths:
                        new_ds[d.num_operands].add_path(*p.indices, c=p.coefficients)

        return cls(new_ds.values(), new_operands)

    def symmetrize_operands(self) -> EquivariantTensorProduct:
        """Symmetrize the operands of the ETP."""
        new_ds = []
        for d in self.ds:
            new_ds.append(
                d.symmetrize_operands(range(self.num_inputs - 1, d.num_operands - 1))
            )
        return EquivariantTensorProduct(new_ds, self.operands)

    def sort_indices_for_identical_operands(self) -> EquivariantTensorProduct:
        """Sort the indices for identical operands."""
        new_ds = []
        for d in self.ds:
            new_ds.append(
                d.sort_indices_for_identical_operands(
                    range(self.num_inputs - 1, d.num_operands - 1)
                )
            )
        return EquivariantTensorProduct(new_ds, self.operands)



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools

import numpy as np
import pytest

import cuequivariance as cue


@pytest.mark.parametrize("Group", [cue.SU2, cue.SO3, cue.O3])
def test_algebra(Group):
    # [X_i, X_j] = A_ijk X_k

    for r in itertools.islice(Group.iterator(), 6):
        xx = np.einsum("iuv,jvw->ijuw", r.X, r.X)
        term1 = xx - np.swapaxes(xx, 0, 1)
        term2 = np.einsum("ijk,kuv->ijuv", r.A, r.X)

        np.testing.assert_allclose(term1, term2, atol=1e-10, rtol=0)



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses

import numpy as np


def np_asarray_with_copy(arr):
    """Create a NumPy array with a copy in a version-compatible way."""
    dtype = np.float64
    order = "C"
    try:
        return np.asarray(arr, dtype=dtype, order=order, copy=True)
    except TypeError:
        return np.asarray(arr, dtype=dtype, order=order).copy()


@dataclasses.dataclass(init=False, frozen=True)
class Path:
    """
    A tensor product path links segments from different operands and contains coefficients.

    Attributes:
        indices (tuple of int): One index per operand, pointing to the segment.
        coefficients (np.ndarray): Coefficients of the path.

    Examples:
        >>> Path((12, 44), 2.0)
        op0[12]*op1[44]*2.
    """

    indices: tuple[int, ...]  # One index per operand, pointing to the segment
    coefficients: np.ndarray

    def __init__(self, indices, coefficients):
        super().__setattr__("indices", tuple(int(i) for i in indices))
        super().__setattr__("coefficients", np_asarray_with_copy(coefficients))

    def assert_valid(self):
        """Assert that the path is valid."""
        if not isinstance(self.indices, tuple):
            raise ValueError(f"indices {self.indices} are not valid.")

        if not all(isinstance(i, int) and i >= 0 for i in self.indices):
            raise ValueError(f"indices {self.indices} are not valid.")

        if not self.coefficients.dtype == np.float64:
            raise ValueError(f"coefficients {self.coefficients} are not valid.")

    def __repr__(self) -> str:
        indices_txt = "*".join(f"op{i}[{s}]" for i, s in enumerate(self.indices))
        txt = np.array2string(
            self.coefficients, separator=" ", precision=2, suppress_small=True
        )
        if "\n" not in txt:
            return f"{indices_txt}*{txt}"
        return f"{indices_txt}*c c.shape={self.coefficients.shape} c.nnz={np.count_nonzero(self.coefficients)}"

    @property
    def num_operands(self) -> int:
        """The number of operands."""
        return len(self.indices)

    def __hash__(self) -> int:
        return hash((self.indices, self.coefficients.tobytes()))

    def __eq__(self, other: Path) -> bool:
        return self.indices == other.indices and np.array_equal(
            self.coefficients, other.coefficients
        )

    def __lt__(self, other: Path) -> bool:
        k1 = (self.indices, self.coefficients.shape)
        k2 = (other.indices, other.coefficients.shape)
        if k1 != k2:
            return k1 < k2
        return tuple(self.coefficients.flatten()) < tuple(other.coefficients.flatten())

    def permute_operands(self, perm: tuple[int, ...]) -> Path:
        """
        Apply a permutation to the operands.

        Args:
            perm (tuple of int): The permutation of the operands.

        Returns:
            Path: A new path with the operands permuted.
        """
        return Path(
            indices=tuple(self.indices[i] for i in perm),
            coefficients=self.coefficients,
        )

    def move_operand(self, operand: int, new_index: int) -> Path:
        """Move an operand to a new position."""
        perm = list(range(self.num_operands))
        perm.remove(operand)
        perm.insert(new_index, operand)
        return self.permute_operands(perm)

    def move_operand_first(self, operand: int) -> Path:
        """Move an operand to the first position."""
        return self.move_operand(operand, 0)

    def move_operand_last(self, operand: int) -> Path:
        """Move an operand to the last position."""
        return self.move_operand(operand, self.num_operands - 1)



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np

import cuequivariance as cue


def test_elasticity_tensor():
    # A general fourth-rank tensor in 3D has 3^4 = 81 independent components, but the elasticity tensor has at most 21 independent components.
    # source: https://en.wikipedia.org/wiki/Elasticity_tensor
    C = cue.reduced_tensor_product_basis("ijkl=klij=jikl", i=cue.Irreps("O3", "1o"))
    assert C.shape == (3, 3, 3, 3, 21)


def test_symmetric_basis():
    C = cue.reduced_symmetric_tensor_product_basis(cue.Irreps("SU2", "1/2 + 1"), 4)
    _irreps, C = C.irreps, C.array
    assert C.shape == (5, 5, 5, 5, 70)

    # The symmetry is respected
    for perm in itertools.permutations(range(4)):
        np.testing.assert_array_equal(C, np.transpose(C, perm + (4,)))

    # All components are independent
    C = np.reshape(C, (5**4, 70))
    assert np.linalg.matrix_rank(C) == 70



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import itertools
import re
from dataclasses import dataclass
from math import factorial
from typing import Iterator

import numpy as np

from cuequivariance.group_theory.representations import Irrep


# This class is an adaptation of https://github.com/lie-nn/lie-nn/blob/70adebce44e3197ee17f780585c6570d836fc2fe/lie_nn/_src/irreps/su2.py
@dataclass(frozen=True)
class SU2(Irrep):
    r"""Subclass of :class:`Irrep`, irreducible representations of the Lie group :math:`SU(2)`.

    Each representation is labeled by a non-negative half-integer :math:`j`.

    Examples:
        >>> SU2(0)
        0
        >>> SU2(1/2)
        1/2
        >>> SU2(1/2).dim
        2
        >>> SU2.from_string("1")
        1
    """

    j: float

    def __post_init__(self):
        assert isinstance(self.j, (float, int))
        assert self.j >= 0
        assert round(2 * self.j) == 2 * self.j

    @classmethod
    def regexp_pattern(cls) -> re.Pattern:
        return re.compile(r"(\d+|\d+/2)")

    @classmethod
    def from_string(cls, string: str) -> SU2:
        string = string.strip()
        if string.endswith("/2"):
            j = int(string[:-2]) / 2
        else:
            j = int(string)
        return cls(j=j)

    def __repr__(rep: SU2) -> str:
        if round(rep.j) == rep.j:
            return f"{int(rep.j)}"
        return f"{int(2 * rep.j)}/2"

    def __mul__(rep1: SU2, rep2: SU2) -> Iterator[SU2]:
        rep2 = rep1._from(rep2)
        return [
            SU2(j=j / 2)
            for j in range(
                abs(int(2 * rep1.j) - int(2 * rep2.j)),
                int(2 * rep1.j + 2 * rep2.j) + 1,
                2,
            )
        ]

    @classmethod
    def clebsch_gordan(cls, rep1: SU2, rep2: SU2, rep3: SU2) -> np.ndarray:
        # return an array of shape ``(number_of_paths, rep1.dim, rep2.dim, rep3.dim)``
        rep1, rep2, rep3 = cls._from(rep1), cls._from(rep2), cls._from(rep3)

        if rep3 in rep1 * rep2:
            return clebsch_gordanSU2mat(rep1.j, rep2.j, rep3.j)[None]
        else:
            return np.zeros((0, rep1.dim, rep2.dim, rep3.dim))

    @property
    def dim(rep: SU2) -> int:
        return int(2 * rep.j + 1)

    def is_scalar(rep: SU2) -> bool:
        return rep.j == 0

    def __lt__(rep1: SU2, rep2: SU2) -> bool:
        rep2 = rep1._from(rep2)
        return rep1.j < rep2.j

    @classmethod
    def iterator(cls) -> Iterator[SU2]:
        for j in itertools.count(0):
            yield SU2(j=j / 2)

    def discrete_generators(rep: SU2) -> np.ndarray:
        return np.zeros((0, rep.dim, rep.dim))

    def continuous_generators(rep: SU2) -> np.ndarray:
        j = rep.j
        m = np.arange(-j, j)
        raising = np.diag(-np.sqrt(j * (j + 1) - m * (m + 1)), k=-1)

        m = np.arange(-j + 1, j + 1)
        lowering = np.diag(np.sqrt(j * (j + 1) - m * (m - 1)), k=1)

        m = np.arange(-j, j + 1)
        return np.stack(
            [
                0.5 * (raising + lowering),  # x (usually)
                np.diag(1j * m),  # z (usually)
                -0.5j * (raising - lowering),  # -y (usually)
            ],
            axis=0,
        )

    def algebra(rep=None) -> np.ndarray:
        # [X_i, X_j] = A_ijk X_k
        return np.array(
            [
                [
                    [0, 0, 0],
                    [0, 0, 1],
                    [0, -1, 0],
                ],
                [
                    [0, 0, -1],
                    [0, 0, 0],
                    [1, 0, 0],
                ],
                [
                    [0, 1, 0],
                    [-1, 0, 0],
                    [0, 0, 0.0],
                ],
            ]
        )


# Taken from http://qutip.org/docs/3.1.0/modules/qutip/utilities.html

# This file is part of QuTiP: Quantum Toolbox in Python.
#
#    Copyright (c) 2011 and later, Paul D. Nation and Robert J. Johansson.
#    All rights reserved.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions are
#    met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of the QuTiP: Quantum Toolbox in Python nor the names
#       of its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
#    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
###############################################################################


def clebsch_gordanSU2mat(j1, j2, j3):
    """Calculates the Clebsch-Gordon matrix
    for SU(2) coupling j1 and j2 to give j3.
    Parameters
    ----------
    j1 : float
        Total angular momentum 1.
    j2 : float
        Total angular momentum 2.
    j3 : float
        Total angular momentum 3.
    Returns
    -------
    cg_matrix : numpy.array
        Requested Clebsch-Gordan matrix.
    """
    assert isinstance(j1, (int, float))
    assert isinstance(j2, (int, float))
    assert isinstance(j3, (int, float))
    mat = np.zeros((int(2 * j1 + 1), int(2 * j2 + 1), int(2 * j3 + 1)))
    if int(2 * j3) in range(int(2 * abs(j1 - j2)), int(2 * (j1 + j2)) + 1, 2):
        for m1 in (x / 2 for x in range(-int(2 * j1), int(2 * j1) + 1, 2)):
            for m2 in (x / 2 for x in range(-int(2 * j2), int(2 * j2) + 1, 2)):
                if abs(m1 + m2) <= j3:
                    mat[int(j1 + m1), int(j2 + m2), int(j3 + m1 + m2)] = (
                        clebsch_gordanSU2coeffs((j1, m1), (j2, m2), (j3, m1 + m2))
                    )
    # in e3nn it was divided by np.sqrt(2 * j3 + 1) to make the CG symmetric wrt j1,j2,j3
    # but if we say that j3 is always the output, it makes sense to not divide
    return mat  # / np.sqrt(2 * j3 + 1)


def clebsch_gordanSU2coeffs(idx1, idx2, idx3):
    """Calculates the Clebsch-Gordon coefficient
    for SU(2) coupling (j1,m1) and (j2,m2) to give (j3,m3).
    Parameters
    ----------
    j1 : float
        Total angular momentum 1.
    j2 : float
        Total angular momentum 2.
    j3 : float
        Total angular momentum 3.
    m1 : float
        z-component of angular momentum 1.
    m2 : float
        z-component of angular momentum 2.
    m3 : float
        z-component of angular momentum 3.
    Returns
    -------
    cg_coeff : float
        Requested Clebsch-Gordan coefficient.
    """
    from fractions import Fraction

    j1, m1 = idx1
    j2, m2 = idx2
    j3, m3 = idx3

    j1 = Fraction(int(2 * j1), 2)
    j2 = Fraction(int(2 * j2), 2)
    j3 = Fraction(int(2 * j3), 2)
    m1 = Fraction(int(2 * m1), 2)
    m2 = Fraction(int(2 * m2), 2)
    m3 = Fraction(int(2 * m3), 2)

    if m3 != m1 + m2:
        return 0
    vmin = int(max([-j1 + j2 + m3, -j1 + m1, 0]))
    vmax = int(min([j2 + j3 + m1, j3 - j1 + j2, j3 + m3]))

    def f(n):
        assert n == round(n)
        return factorial(round(n))

    C = (2 * j3 + 1) * Fraction(
        f(j3 + j1 - j2) * f(j3 - j1 + j2) * f(j1 + j2 - j3) * f(j3 + m3) * f(j3 - m3),
        f(j1 + j2 + j3 + 1) * f(j1 - m1) * f(j1 + m1) * f(j2 - m2) * f(j2 + m2),
    )

    S = 0
    for v in range(vmin, vmax + 1):
        S += (-1) ** (v + j2 + m2) * Fraction(
            f(j2 + j3 + m1 - v) * f(j1 - m1 + v),
            f(v) * f(j3 - j1 + j2 - v) * f(j3 + m3 - v) * f(v + j1 - j2 - m3),
        )

    if S > 0:
        return (C * S**2) ** 0.5
    else:
        return -((C * S**2) ** 0.5)



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import math
from typing import Optional, Union

import numpy as np

import cuequivariance as cue


def compute_last_operand(
    descriptor: cue.SegmentedTensorProduct,
    *inputs: np.ndarray,
    segment_axes: Union[int, list[int]] = -1,
    dtype: Optional[np.dtype] = None,
) -> np.ndarray:
    r"""
    Compute the last operand of a segmented tensor product.

    Args:
        descriptor (SegmentedTensorProduct): The descriptor of the segmented tensor product.
        inputs (np.ndarray): The operands of the segmented tensor product.
        segment_axes (int or list of int, optional): The axes along which to segment the inputs, by default -1.

    Returns:
        np.ndarray: The last operand of the segmented tensor product.
    """
    if len(inputs) != descriptor.num_operands - 1:
        raise ValueError(
            f"Expected {descriptor.num_operands - 1} inputs, got {len(inputs)}."
        )

    if isinstance(segment_axes, int):
        segment_axes = [segment_axes] * descriptor.num_operands

    segment_axes = [k % input.ndim for k, input in zip(segment_axes, inputs)] + [
        segment_axes[-1]
    ]
    assert len(segment_axes) == descriptor.num_operands

    for i, (operand, input, k) in enumerate(
        zip(descriptor.operands[:-1], inputs, segment_axes)
    ):
        if operand.size != input.shape[k]:
            raise ValueError(
                f"Expected operand {i} to have size {operand.size}, got {input.shape[k]}."
            )

    batch_shapes = [
        input.shape[:k] + input.shape[k + 1 :] for input, k in zip(inputs, segment_axes)
    ]
    output_batch_shape = np.broadcast_shapes(*batch_shapes)

    segment_slices = [operand.segment_slices() for operand in descriptor.operands]
    return primitive_compute_last_operand(
        descriptor.subscripts.operands,
        descriptor.subscripts.coefficients,
        [operand.segments for operand in descriptor.operands],
        output_batch_shape,
        [[slice_.start for slice_ in segment] for segment in segment_slices],
        [
            [slice_.stop - slice_.start for slice_ in segment]
            for segment in segment_slices
        ],
        [path.indices for path in descriptor.paths],
        [path.coefficients for path in descriptor.paths],
        segment_axes,
        dtype or np.result_type(*inputs),
        *inputs,
    )


def primitive_compute_last_operand(
    operand_subscripts: list[str],  # ['uvw', 'iu', 'jv', 'kw']
    coefficient_subscripts: str,  # 'ijk'
    segment_shapes: list[list[tuple[int, ...]]],  # segment_shapes[oid][sid] = shape
    output_batch_shape: tuple[int, ...],
    segment_offsets: list[list[slice]],  # segment_offsets[oid][sid] = offset
    segment_sizes: list[list[int]],  # segment_sizes[oid][sid] = size
    indices: list[tuple[int, ...]],  # indices[pid][oid] = sid
    coefficients: Union[list[np.ndarray], np.ndarray],  # coefficients[pid] = coeff
    segment_axes: list[int],  # segment_axes[oid] = axis
    dtype,
    *inputs: np.ndarray,
) -> np.ndarray:
    num_operands = len(operand_subscripts)
    inputs_subscripts = ",".join("..." + s for s in operand_subscripts[:-1])
    if inputs_subscripts:
        formula = (
            f"{coefficient_subscripts},{inputs_subscripts}->...{operand_subscripts[-1]}"
        )
    else:
        formula = f"{coefficient_subscripts}->{operand_subscripts[-1]}"

    output = np.zeros(
        output_batch_shape + (sum(segment_sizes[-1]),),
        dtype=dtype,
    )

    inputs = [np.moveaxis(input, k, -1) for input, k in zip(inputs, segment_axes)]

    for idxx, coeff in zip(indices, coefficients):
        slices = [
            slice(
                segment_offsets[oid][idxx[oid]],
                segment_offsets[oid][idxx[oid]] + segment_sizes[oid][idxx[oid]],
            )
            for oid in range(num_operands)
        ]
        output[..., slices[-1]] += np.reshape(
            np.einsum(
                formula,
                coeff,
                *[
                    np.reshape(
                        inputs[oid][..., slices[oid]],
                        inputs[oid].shape[:-1] + segment_shapes[oid][idxx[oid]],
                    )
                    for oid in range(num_operands - 1)
                ],
            ),
            output_batch_shape + (math.prod(segment_shapes[-1][idxx[-1]]),),
        )

    return np.moveaxis(output, -1, segment_axes[-1])



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import wraps
from typing import Optional, Type, Union

import cuequivariance as cue  # noqa: F401
import cuequivariance.group_theory.irreps_array as irreps_array
from cuequivariance.group_theory import Irrep
from cuequivariance.group_theory.irreps_array.context_irrep_class import (
    pop_irrep_scope,
    push_irrep_scope,
)
from cuequivariance.group_theory.irreps_array.context_layout import (
    pop_layout_scope,
    push_layout_scope,
)


class assume:
    """
    ``assume`` is a context manager or decorator to assume the irrep class and layout for a block of code or a function.

    Examples:
        As a context manager:

        >>> with cue.assume(cue.SO3, cue.mul_ir):
        ...     rep = cue.IrrepsAndLayout("2x1")
        >>> rep.irreps
        2x1
        >>> rep.layout
        (mul,irrep)

        As a decorator:

        >>> @cue.assume(cue.SO3, cue.mul_ir)
        ... def foo():
        ...     return cue.IrrepsAndLayout("2x1")
        >>> assert foo() == rep
    """

    def __init__(
        self,
        irrep_class: Optional[Union[str, Type[Irrep]]] = None,
        layout: Optional[irreps_array.IrrepsLayout] = None,
    ):
        if isinstance(irrep_class, irreps_array.IrrepsLayout) and layout is None:
            irrep_class, layout = None, irrep_class

        self.irrep_class = irrep_class
        self.layout = layout

    def __enter__(self):
        self.old_irrep_class = push_irrep_scope(self.irrep_class)
        self.old_layout = push_layout_scope(self.layout)
        return self

    def __exit__(self, *exc):
        pop_irrep_scope(self.old_irrep_class)
        pop_layout_scope(self.old_layout)
        return False

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return wrapper



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .transposition import transpose
from .irreps_tp import (
    full_tensor_product,
    fully_connected_tensor_product,
    channelwise_tensor_product,
    elementwise_tensor_product,
    linear,
)
from .symmetric_contractions import symmetric_contraction
from .rotations import (
    fixed_axis_angle_rotation,
    y_rotation,
    x_rotation,
    xy_rotation,
    yx_rotation,
    yxy_rotation,
    inversion,
)
from .spherical_harmonics_ import sympy_spherical_harmonics, spherical_harmonics

__all__ = [
    "transpose",
    "full_tensor_product",
    "fully_connected_tensor_product",
    "channelwise_tensor_product",
    "elementwise_tensor_product",
    "linear",
    "symmetric_contraction",
    "fixed_axis_angle_rotation",
    "y_rotation",
    "x_rotation",
    "xy_rotation",
    "yx_rotation",
    "yxy_rotation",
    "inversion",
    "sympy_spherical_harmonics",
    "spherical_harmonics",
]



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from enum import Enum, auto
from typing import Union

import cuequivariance as cue


class IrrepsLayout(Enum):
    """
    Enum for the possible data layouts.

    Attributes:
        mul_ir: Multiplicity first, then irreducible representation.
            This layout corresponds to the layout used in the library `e3nn`.
        ir_mul: Irreducible representation first, then multiplicity.
            This layout differs from the one used in `e3nn` but can be more convenient in some cases.

    Examples:
        >>> cue.mul_ir
        (mul,irrep)

        >>> cue.ir_mul
        (irrep,mul)

    .. rubric:: Methods
    """

    mul_ir = auto()
    ir_mul = auto()

    def shape(
        self, mulir: Union[cue.MulIrrep, tuple[int, cue.Irrep]]
    ) -> tuple[int, int]:
        """The shape of the tensor for the given layout.

        Examples:
            >>> mulir = cue.MulIrrep(32, cue.O3(2, -1))
            >>> mulir
            32x2o

            >>> cue.mul_ir.shape(mulir)
            (32, 5)

            >>> cue.ir_mul.shape(mulir)
            (5, 32)
        """
        mul, ir = mulir
        if self == IrrepsLayout.mul_ir:
            return (mul, ir.dim)
        return (ir.dim, mul)

    def __repr__(self) -> str:
        if self == IrrepsLayout.mul_ir:
            return "(mul,irrep)"
        if self == IrrepsLayout.ir_mul:
            return "(irrep,mul)"

    def __str__(self) -> str:
        return self.__repr__()

    @staticmethod
    def as_layout(layout: Union[str, IrrepsLayout, None]) -> IrrepsLayout:
        if isinstance(layout, IrrepsLayout):
            return layout
        if layout is None:
            return cue.get_layout_scope()
        try:
            return cue.IrrepsLayout[layout]
        except KeyError:
            raise ValueError(f"Invalid layout {layout}")


mul_ir = IrrepsLayout.mul_ir
ir_mul = IrrepsLayout.ir_mul



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses
import itertools
from typing import Any, Sequence

import numpy as np

import cuequivariance as cue


def stable_unique(xs: Sequence[Any]) -> Sequence[Any]:
    seen = set()
    gen = (x for x in xs if not (x in seen or seen.add(x)))

    ty = type(xs)
    if isinstance(xs, str):
        return ty("".join(gen))
    return ty(gen)


def dot(
    x: cue.SegmentedTensorProduct,
    y: cue.SegmentedTensorProduct,
    *contraction: tuple[int, int],
) -> cue.SegmentedTensorProduct:
    """
    Compute the dot product of two segmented tensor products.

    Args:
        x: The first segmented tensor product.
        y: The second segmented tensor product.
        contraction: A tuple of two integers representing the indices of the operands to contract.

    Returns:
        The segmented tensor product resulting from the dot product.
    """
    for oidx, oidy in contraction:
        if x.operands_and_subscripts[oidx] != y.operands_and_subscripts[oidy]:
            raise ValueError("Operands to contract must be the same.")

    x_keep = [
        oid for oid in range(x.num_operands) if all(oid != i for i, _ in contraction)
    ]
    y_keep = [
        oid for oid in range(y.num_operands) if all(oid != j for _, j in contraction)
    ]

    d = cue.SegmentedTensorProduct(
        coefficient_subscripts=stable_unique(
            x.coefficient_subscripts + y.coefficient_subscripts
        )
    )
    d = dataclasses.replace(
        d,
        operands_and_subscripts=[x.operands_and_subscripts[i] for i in x_keep]
        + [y.operands_and_subscripts[i] for i in y_keep],
    )

    formula = f"{x.coefficient_subscripts} , {y.coefficient_subscripts} -> {d.coefficient_subscripts}"

    if len(contraction) == 0:
        for pathx, pathy in itertools.product(x.paths, y.paths):
            d.add_path(
                *pathx.indices,
                *pathy.indices,
                c=np.einsum(formula, pathx.coefficients, pathy.coefficients),
            )
        return d

    oidx, oidy = contraction[0]
    operand = x.operands[oidx]  # = y.operands[oidy]

    x, y = x.sort_paths(oidx), y.sort_paths(oidy)

    cx = x.compressed_path_segment(oidx)
    cy = y.compressed_path_segment(oidy)
    for sid in range(operand.num_segments):
        for pathx, pathy in itertools.product(
            x.paths[cx[sid] : cx[sid + 1]], y.paths[cy[sid] : cy[sid + 1]]
        ):
            assert pathx.indices[oidx] == pathy.indices[oidy]
            if all(
                pathx.indices[oidx_] == pathy.indices[oidy_]
                for oidx_, oidy_ in contraction
            ):
                d.add_path(
                    *[pathx.indices[i] for i in x_keep],
                    *[pathy.indices[i] for i in y_keep],
                    c=np.einsum(formula, pathx.coefficients, pathy.coefficients),
                )

    return d


def trace(
    d: cue.SegmentedTensorProduct, *contraction: tuple[int, int]
) -> cue.SegmentedTensorProduct:
    """
    Compute the trace of a segmented tensor product.

    Args:
        d: The segmented tensor product.
        contraction: A tuple of two integers representing the indices of the operands to contract.

    Returns:
        The segmented tensor product resulting from the trace.
    """
    for oidx, oidy in contraction:
        if d.operands[oidx].segments != d.operands[oidy].segments:
            raise ValueError("Operands to contract must be the same.")

    keep = [
        oid
        for oid in range(d.num_operands)
        if all(oid not in (i, j) for i, j in contraction)
    ]
    mapping = {
        chj: chi
        for i, j in contraction
        for chi, chj in zip(d.subscripts.operands[i], d.subscripts.operands[j])
    }
    f = lambda subscripts: "".join(mapping.get(ch, ch) for ch in subscripts)  # noqa

    coefficients_subscripts_renamed = f(d.coefficient_subscripts)
    coefficients_subscripts_compressed = stable_unique(coefficients_subscripts_renamed)

    dout = cue.SegmentedTensorProduct(
        coefficient_subscripts=coefficients_subscripts_compressed,
        operands_and_subscripts=[
            (ope, f(ss)) for ope, ss in (d.operands_and_subscripts[i] for i in keep)
        ],
    )

    formula = (
        f"{coefficients_subscripts_renamed} -> {coefficients_subscripts_compressed}"
    )
    for path in d.paths:
        if all(path.indices[i] == path.indices[j] for i, j in contraction):
            dout.add_path(
                *[path.indices[i] for i in keep],
                c=np.einsum(formula, path.coefficients),
            )

    return dout



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import numpy as np
import scipy.linalg


# This class is adapted from https://github.com/lie-nn/lie-nn/blob/70adebce44e3197ee17f780585c6570d836fc2fe/lie_nn/_src/rep.py
class Rep:
    r"""Abstract Class, Representation of a Lie group.

    ``Rep`` abstractly defines what a group representation is and how it can be used.
    """

    @property
    def lie_dim(self) -> int:
        """Dimension of the Lie algebra

        Returns:
            int: The dimension of the Lie algebra.
        """
        A = self.algebra()
        d = A.shape[0]
        return d

    @property
    def dim(self) -> int:
        """Dimension of the representation

        Returns:
            int: The dimension of the representation.
        """
        return self._dim()

    def _dim(self) -> int:
        X = self.continuous_generators()
        d = X.shape[1]
        return d

    @property
    def A(self) -> np.ndarray:
        """
        Algebra of the Lie group, ``(lie_dim, lie_dim, lie_dim)``

        See Also:
            :meth:`algebra`
        """
        return self.algebra()

    @property
    def X(self) -> np.ndarray:
        """
        Generators of the representation, ``(lie_dim, dim, dim)``

        See Also:
            :meth:`continuous_generators`
        """
        return self.continuous_generators()

    @property
    def H(self) -> np.ndarray:
        """
        Discrete generators of the representation, ``(len(H), dim, dim)``

        See Also:
            :meth:`discrete_generators`
        """
        return self.discrete_generators()

    def algebra(self) -> np.ndarray:
        """Algebra of the Lie group

        The algebra of the Lie group is defined by the following equation:

        .. math::

            [X_i, X_j] = A_{ijk} X_k

        where :math:`X_i` are the continuous generators and :math:`A_{ijk}` is the algebra.

        Returns:
            np.ndarray: An array of shape ``(lie_dim, lie_dim, lie_dim)``.
        """
        raise NotImplementedError  # pragma: no cover

    def continuous_generators(self) -> np.ndarray:
        r"""Generators of the representation

        The generators of the representation are defined by the following equation:

        .. math::

            \rho(\alpha) = \exp\left(\alpha_i X_i\right)

        Where :math:`\rho(\alpha)` is the representation of the group element
        corresponding to the parameter :math:`\alpha` and :math:`X_i` are the
        (continuous) generators of the representation, each of shape ``(dim, dim)``.

        Returns:
            np.ndarray: An array of shape ``(lie_dim, dim, dim)``.
        """
        raise NotImplementedError  # pragma: no cover

    def discrete_generators(self) -> np.ndarray:
        r"""Discrete generators of the representation

        .. math::

            \rho(n) = H^n

        Returns:
            np.ndarray: An array of shape ``(len(H), dim, dim)``.
        """
        raise NotImplementedError  # pragma: no cover

    def trivial(self) -> Rep:
        """
        Create a trivial representation from the same group as self
        """
        raise NotImplementedError  # pragma: no cover

    def exp_map(
        self, continuous_params: np.ndarray, discrete_params: np.ndarray
    ) -> np.ndarray:
        """Exponential map of the representation

        Args:
            continuous_params (np.ndarray): An array of shape ``(lie_dim,)``.
            discrete_params (np.ndarray): An array of shape ``(len(H),)``.

        Returns:
            np.ndarray: An matrix of shape ``(dim, dim)``.
        """
        output = scipy.linalg.expm(
            np.einsum("a,aij->ij", continuous_params, self.continuous_generators())
        )
        for k, h in reversed(list(zip(discrete_params, self.discrete_generators()))):
            output = np.linalg.matrix_power(h, k) @ output
        return output

    def is_scalar(self) -> bool:
        """Check if the representation is scalar (acting as the identity)"""
        return np.all(self.X == 0.0) and np.all(self.H == np.eye(self.dim))

    def is_trivial(self) -> bool:
        """Check if the representation is trivial (scalar of dimension 1)"""
        return self.dim == 1 and self.is_scalar()

    def __eq__(self, other: Rep) -> bool:
        return (
            self.dim == other.dim
            and np.allclose(self.A, other.A)
            and np.allclose(self.H, other.H)
            and np.allclose(self.X, other.X)
        )

    def __repr__(self) -> str:
        return f"Rep(dim={self.dim}, lie_dim={self.lie_dim}, len(H)={len(self.H)})"



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
import pytest

import cuequivariance as cue


def test_transpose():
    cue.descriptors.transpose(cue.Irreps(cue.O3, "3x1e"), cue.mul_ir, cue.ir_mul)


def test_commutativity_squeeze_flatten():
    irreps1 = cue.Irreps("O3", "32x0e + 32x1o")
    irreps2 = cue.Irreps("O3", "1x0e + 1x1o")
    irreps3 = cue.Irreps("O3", "32x0e + 32x1o")

    poly = cue.descriptors.fully_connected_tensor_product(irreps1, irreps2, irreps3)
    assert (
        poly.squeeze_modes().flatten_coefficient_modes()
        == poly.flatten_coefficient_modes().squeeze_modes()
    )

    poly = cue.descriptors.full_tensor_product(irreps1, irreps2, irreps3)
    assert (
        poly.squeeze_modes().flatten_coefficient_modes()
        == poly.flatten_coefficient_modes().squeeze_modes()
    )

    poly = cue.descriptors.channelwise_tensor_product(irreps1, irreps2, irreps3)
    assert (
        poly.squeeze_modes().flatten_coefficient_modes()
        == poly.flatten_coefficient_modes().squeeze_modes()
    )

    poly = cue.descriptors.linear(irreps1, irreps2)
    assert (
        poly.squeeze_modes().flatten_coefficient_modes()
        == poly.flatten_coefficient_modes().squeeze_modes()
    )


@pytest.mark.parametrize("ell", [1, 2, 3, 4])
def test_spherical_harmonics(ell: int):
    poly = cue.descriptors.spherical_harmonics(cue.SO3(1), [ell])

    vec = np.random.randn(3)
    axis = np.random.randn(3)
    angle = np.random.rand()

    [yl] = poly(vec)

    R = cue.SO3(1).rotation(axis, angle)
    Rl = cue.SO3(ell).rotation(axis, angle)

    [yl1] = poly(R @ vec)
    yl2 = Rl @ yl

    np.testing.assert_allclose(yl1, yl2, rtol=1e-06, atol=1e-13)
    np.testing.assert_allclose(np.sum(yl**2), (2 * ell + 1) * np.sum(vec**2) ** ell)


@pytest.mark.parametrize("ell", [0, 1, 2, 3, 4])
def test_y_rotation(ell: int):
    alpha = 0.3
    beta = 0.4
    gamma = -0.5

    irrep = cue.SO3(ell)
    poly = cue.descriptors.yxy_rotation(cue.Irreps("SO3", [irrep]))

    def enc(th: float):
        m = np.arange(1, ell + 1)
        c = np.cos(m * th)
        s = np.sin(m * th)
        return np.concatenate([c[::-1], [1.0], s])

    x = np.random.randn(irrep.dim)
    [y1] = poly(enc(gamma), enc(beta), enc(alpha), x)

    A = irrep.rotation(np.array([0.0, 1.0, 0.0]), alpha)
    B = irrep.rotation(np.array([1.0, 0.0, 0.0]), beta)
    C = irrep.rotation(np.array([0.0, 1.0, 0.0]), gamma)
    y2 = A @ B @ C @ x

    np.testing.assert_allclose(y1, y2)


def test_elementwise_tensor_product():
    irreps1 = cue.Irreps("O3", "8x0e + 7x1o + 4x2e")
    irreps2 = cue.Irreps("O3", "8x0e + 7x1o + 4x0e")
    irreps3 = cue.Irreps("O3", "0e + 1o + 1e")
    poly = cue.descriptors.elementwise_tensor_product(irreps1, irreps2, irreps3)
    assert (
        poly.squeeze_modes().flatten_coefficient_modes()
        == poly.flatten_coefficient_modes().squeeze_modes()
    )
    assert poly.num_inputs == 2
    assert poly.num_outputs == 1


def test_symmetric_contraction():
    irreps_in = 16 * cue.Irreps("SO3", "0 + 1 + 2")
    irreps_out = 16 * cue.Irreps("SO3", "0 + 1")
    poly = cue.descriptors.symmetric_contraction(irreps_in, irreps_out, [0, 1, 2, 3])
    assert poly.num_inputs == 2
    assert poly.num_outputs == 1

    [_, _, _, (_, d)] = poly.polynomial.operations
    assert d.num_paths == 437

    poly = poly.polynomial.unsymmetrize_for_identical_operands()
    [_, _, _, (_, d)] = poly.operations
    assert d.num_paths == 105
